/**
 * Copyright 2019 Kitchener-Waterloo Youth Robotics
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not
 * use this file except in compliance with the License. You may obtain a copy of
 * the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <frc/DriverStation.h>
#include <frc/Timer.h>
#include <frc/livewindow/LiveWindow.h>

#include "BaseRobot.h"
#include "Settings.h"
#include "StateMachines.h"
#include "Telemetry.h"

using namespace Rebels;

#define TELEMETRY_CATEGORY "Robot"

void BaseRobot::Internal::RobotInit() {
  LOG_INFO(TELEMETRY_CATEGORY, "=== ROBOT INITIALIZATION STARTED ===");
  LOG_INFO(TELEMETRY_CATEGORY, fmt::format("Update cycle is every {:f}ms", GetPeriod().convert<units::millisecond>().value()));

  sched_param sch_params;
  int policy;
  if (pthread_getschedparam(pthread_self(), &policy, &sch_params)) {
    LOG_ERROR(TELEMETRY_CATEGORY, fmt::format("Failed to get thread priority: {}", strerror(errno)));
  }
  else {
    LOG_INFO(TELEMETRY_CATEGORY, fmt::format("Main thread has policy {} with priority={}", policy, sch_params.sched_priority));
  }

  srand(time(0));

  // Disable LiveWindow, helps with some PDP timeouts.
  frc::LiveWindow::DisableAllTelemetry();
  parent->RobotInit();
  LOG_INFO(TELEMETRY_CATEGORY, "=== ROBOT INITIALIZATION ENDED ===");
}

void BaseRobot::Internal::RobotPeriodic() {
  // Update all registered state machines
  StateMachines::UpdateAll();
  parent->RobotPeriodic();
}

void BaseRobot::Internal::DisabledInit() {
  LOG_INFO(TELEMETRY_CATEGORY, fmt::format("=== DISABLED INITIALIZATION STARTED === (enabled={}, inmatch={})", wasEnabled, wasInMatch));
  Settings::EnableWrites();

  // The default FMS implementation will go from: Auto->Disabled->Teleop->Disabled
  // We don't want to reset all our systems during a match.
  if (wasEnabled && !wasInMatch) {
    parent->DisabledInit();
    enabledInitCalled = false;
  }

  disableTime = frc::Timer::GetFPGATimestamp();
  wasEnabled = false;
  LOG_INFO(TELEMETRY_CATEGORY, "=== DISABLED INITIALIZATION ENDED ===");
}

void BaseRobot::Internal::DisabledPeriodic() {
  parent->DisabledPeriodic();
}

void BaseRobot::Internal::EnabledInit() {
  if (enabledInitCalled) {
    return;
  }

  enabledInitCalled = true;

  LOG_INFO(TELEMETRY_CATEGORY, "=== ENABLED INITIALIZATION STARTED ===");
  Settings::DisableWrites();
  parent->EnabledInit();
  LOG_INFO(TELEMETRY_CATEGORY, "=== ENABLED INITIALIZATION ENDED ===");
}

void BaseRobot::Internal::AutonomousInit() {
  EnabledInit();
  LOG_INFO(TELEMETRY_CATEGORY, "=== AUTONOMOUS INITIALIZATION STARTED ===");
  wasEnabled = true;
  wasInMatch = frc::DriverStation::GetMatchTime() > -1;
  parent->AutonomousInit();
  LOG_INFO(TELEMETRY_CATEGORY, "=== AUTONOMOUS INITIALIZATION ENDED ===");
}

void BaseRobot::Internal::AutonomousPeriodic() {
  parent->AutonomousPeriodic();
}

void BaseRobot::Internal::TeleopInit() {
  EnabledInit();
  LOG_INFO(TELEMETRY_CATEGORY, "=== TELEOP INITIALIZATION STARTED ===");
  wasEnabled = true;
  wasInMatch = frc::DriverStation::GetMatchTime() > -1;
  parent->TeleopInit();
  LOG_INFO(TELEMETRY_CATEGORY, "=== TELEOP INITIALIZATION ENDED ===");
}

void BaseRobot::Internal::TeleopPeriodic() {
  parent->TeleopPeriodic();
}
