/**
 * Copyright 2019 Kitchener-Waterloo Youth Robotics
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not
 * use this file except in compliance with the License. You may obtain a copy of
 * the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <hal/CAN.h>

#include "Constants.h"
#include "Drive.h"
#include "Telemetry.h"
#include "Settings.h"

using namespace Rebels;

#define DOUBLE_SQUARE_TURNS
#define TELEMETRY_CATEGORY "Drive"

REGISTER_BOOL(DriveVerboseDashboard, "Drive/Verbose", true)
REGISTER_DOUBLE(DriveOpenLoopRamp, "Drive/Ramp", 0.5)
REGISTER_DOUBLE(DriveDeadband, "Drive/Deadband", 0.05)
REGISTER_STRING(DriveOwner, "Drive/Owner", "None")

PrivateDrive::PrivateDrive() : StateMachine("PrivateDrive", ShifterStrings) {
  AddTransition(SHIFTER_INIT, SHIFTER_SHIFTING_LOW);
  AddTransition(SHIFTER_INIT, SHIFTER_SHIFTING_HIGH);
  AddTransition(SHIFTER_HIGH_GEAR, SHIFTER_SHIFTING_LOW);
  AddTransition(SHIFTER_HIGH_GEAR, SHIFTER_SHIFTING_HIGH);
  AddTransition(SHIFTER_LOW_GEAR, SHIFTER_SHIFTING_HIGH);
  AddTransition(SHIFTER_LOW_GEAR, SHIFTER_SHIFTING_LOW);
  AddTransition(SHIFTER_SHIFTING_LOW, SHIFTER_LOW_GEAR);
  AddTransition(SHIFTER_SHIFTING_HIGH, SHIFTER_HIGH_GEAR);

  imu.reset(Imu::Get());

#ifdef ENABLE_DRIVE
  LOG_INFO(TELEMETRY_CATEGORY, "Setting up the drive motors");
  leftMotorLeader.RestoreFactoryDefaults();
  rightMotorLeader.RestoreFactoryDefaults();
  leftMotorFollower.RestoreFactoryDefaults();
  rightMotorFollower.RestoreFactoryDefaults();

  leftMotorLeader.SetInverted(true);
  rightMotorLeader.SetInverted(false);

  leftMotorLeader.SetOpenLoopRampRate(GetDriveOpenLoopRamp());
  rightMotorLeader.SetOpenLoopRampRate(GetDriveOpenLoopRamp());

  leftMotorFollower.Follow(leftMotorLeader, false);
  rightMotorFollower.Follow(rightMotorLeader, false);

  diffDrive.SetDeadband(GetDriveDeadband());
  // diffDrive.SetSafetyEnabled(false);
#endif

  SetRotationTranslationFactor(DRIVE_HIGH_GEAR_FACTOR);
  SetCoastMode();
  ResetOdometry();
}

PrivateDrive::~PrivateDrive() {
  if (!owner.empty()) {
    Release(owner);
  }
}

bool PrivateDrive::TransitionTo(ShifterState newState) {
  switch (newState) {
  case SHIFTER_INIT:
    SetRotationTranslationFactor(DRIVE_HIGH_GEAR_FACTOR);
    return true;
  case SHIFTER_LOW_GEAR:
    SetShifterSolenoid(frc::DoubleSolenoid::kOff);
    SetRotationTranslationFactor(DRIVE_LOW_GEAR_FACTOR);
    return true;
  case SHIFTER_SHIFTING_LOW:
    SetShifterSolenoid(frc::DoubleSolenoid::kReverse);
    return true;
  case SHIFTER_HIGH_GEAR:
    SetShifterSolenoid(frc::DoubleSolenoid::kOff);
    SetRotationTranslationFactor(DRIVE_HIGH_GEAR_FACTOR);
    return true;
  case SHIFTER_SHIFTING_HIGH:
    SetShifterSolenoid(frc::DoubleSolenoid::kForward);
    return true;
  }
  return false;
}

void PrivateDrive::SetShifterSolenoid(frc::DoubleSolenoid::Value v) {
#ifdef ENABLE_SHIFTERS
  shifterSolenoid.Set(v);
#endif
}

void PrivateDrive::Update() {
  switch (GetCurrentState()) {
  case SHIFTER_INIT:
    ChangeState(SHIFTER_SHIFTING_HIGH);
    break;
  case SHIFTER_LOW_GEAR:
    break;
  case SHIFTER_SHIFTING_LOW:
    if (ElapsedSinceTransition() > 250_ms) {
      ChangeState(SHIFTER_LOW_GEAR);
    }
    break;
  case SHIFTER_HIGH_GEAR:
    break;
  case SHIFTER_SHIFTING_HIGH:
    if (ElapsedSinceTransition() > 250_ms) {
      ChangeState(SHIFTER_HIGH_GEAR);
    }
    break;
  }

#ifdef ENABLE_DRIVE
  odometry.Update(
    imu->GetHeading(),
    leftMotorEncoder.GetPosition() * rotationTranslationFactor,
    rightMotorEncoder.GetPosition() * rotationTranslationFactor);
#endif
}

void PrivateDrive::BackgroundUpdate() {
#ifdef ENABLE_DRIVE
  if (!GetDriveVerboseDashboard()) {
    return;
  }

  {
    int32_t status = 0;
    float percentBusUtilization = 0;
    uint32_t busOffCount = 0;
    uint32_t txFullCount = 0;
    uint32_t receiveErrorCount = 0;
    uint32_t transmitErrorCount = 0;
    HAL_CAN_GetCANStatus(
      &percentBusUtilization,
      &busOffCount,
      &txFullCount,
      &receiveErrorCount,
      &transmitErrorCount,
      &status);

    if (status != 0) {
      LOG_ERROR(TELEMETRY_CATEGORY, fmt::format("Non-successful CAN status: {}", status));
      // TODO: what shall we do here?
      // wpi_setGlobalHALError(status);
    }
    else {
      LOG_VALUE(TELEMETRY_CATEGORY, "CAN bus % utilization", percentBusUtilization);
      LOG_VALUE(TELEMETRY_CATEGORY, "CAN bus OFF errors", (double)busOffCount);
      LOG_VALUE(TELEMETRY_CATEGORY, "CAN bus TX FULL errors", (double)txFullCount);
      LOG_VALUE(TELEMETRY_CATEGORY, "CAN bus RX errors", (double)receiveErrorCount);
      LOG_VALUE(TELEMETRY_CATEGORY, "CAN bus TX errors", (double)transmitErrorCount);
    }
  }

  const auto pose = odometry.GetPose();
  const auto translation = pose.Translation();
  const auto rotation = pose.Rotation();

  LOG_VALUE(TELEMETRY_CATEGORY, "Odometry/Distance (m)", translation.Norm().value());
  LOG_VALUE(TELEMETRY_CATEGORY, "Odometry/X (m)", translation.X().value());
  LOG_VALUE(TELEMETRY_CATEGORY, "Odometry/Y (m)", translation.Y().value());
  LOG_VALUE(TELEMETRY_CATEGORY, "Odometry/Rotation (deg)", rotation.Degrees().value());

  const double leftVelocityRpm = leftMotorEncoder.GetVelocity();
  const double rightVelocityRpm = rightMotorEncoder.GetVelocity();
  const units::meters_per_second_t leftVelocity = (leftVelocityRpm * rotationTranslationFactor) / 60_s;
  const units::meters_per_second_t rightVelocity = (rightVelocityRpm * rotationTranslationFactor) / 60_s;

  LOG_VALUE(TELEMETRY_CATEGORY, "L-Motor/Velocity (RPM)", leftVelocityRpm);
  LOG_VALUE(TELEMETRY_CATEGORY, "R-Motor/Velocity (RPM)", rightVelocityRpm);
  LOG_VALUE(TELEMETRY_CATEGORY, "L-Motor/Velocity (m per sec)", leftVelocity.value());
  LOG_VALUE(TELEMETRY_CATEGORY, "R-Motor/Velocity (m per sec)", rightVelocity.value());
  LOG_VALUE(TELEMETRY_CATEGORY, "L-Motor/Rotations", leftMotorEncoder.GetPosition());
  LOG_VALUE(TELEMETRY_CATEGORY, "R-Motor/Rotations", rightMotorEncoder.GetPosition());

  LOG_VALUE(TELEMETRY_CATEGORY, "L-Motor/Current (A)", leftMotorLeader.GetOutputCurrent());
  LOG_VALUE(TELEMETRY_CATEGORY, "R-Motor/Current (A)", rightMotorLeader.GetOutputCurrent());
  LOG_VALUE(TELEMETRY_CATEGORY, "Motor Current Delta (A)", rightMotorLeader.GetOutputCurrent() - leftMotorLeader.GetOutputCurrent());
  LOG_VALUE(TELEMETRY_CATEGORY, "L-Motor/Voltage (V)", leftMotorLeader.GetBusVoltage());
  LOG_VALUE(TELEMETRY_CATEGORY, "R-Motor/Voltage (V)", rightMotorLeader.GetBusVoltage());
  LOG_VALUE(TELEMETRY_CATEGORY, "L-Motor/Temperature (C)", leftMotorLeader.GetMotorTemperature());
  LOG_VALUE(TELEMETRY_CATEGORY, "R-Motor/Temperature (C)", rightMotorLeader.GetMotorTemperature());

  SetDriveOwner(std::string(GetOwner()));
#endif
}

bool PrivateDrive::IsCoastMode() const {
  return coastMode;
}

bool PrivateDrive::IsBrakeMode() const {
  return !coastMode;
}

void PrivateDrive::SetCoastMode() {
  LOG_INFO(TELEMETRY_CATEGORY, "Enabling coast mode");
  coastMode = true;

#ifdef ENABLE_DRIVE
  leftMotorLeader.SetIdleMode(rev::CANSparkMax::IdleMode::kCoast);
  leftMotorFollower.SetIdleMode(rev::CANSparkMax::IdleMode::kCoast);
  rightMotorLeader.SetIdleMode(rev::CANSparkMax::IdleMode::kCoast);
  rightMotorFollower.SetIdleMode(rev::CANSparkMax::IdleMode::kCoast);
#endif
}

void PrivateDrive::SetBrakeMode() {
  LOG_INFO(TELEMETRY_CATEGORY, "Enabling brake mode");
  coastMode = false;

#ifdef ENABLE_DRIVE
  leftMotorLeader.SetIdleMode(rev::CANSparkMax::IdleMode::kBrake);
  leftMotorFollower.SetIdleMode(rev::CANSparkMax::IdleMode::kBrake);
  rightMotorLeader.SetIdleMode(rev::CANSparkMax::IdleMode::kBrake);
  rightMotorFollower.SetIdleMode(rev::CANSparkMax::IdleMode::kBrake);
#endif
}

void PrivateDrive::SetRotationTranslationFactor(double gearFactor) {
  rotationTranslationFactor = DRIVE_WHEEL_CIRCUMFERENCE / gearFactor;
}

void PrivateDrive::ResetOdometry() {
#ifdef ENABLE_DRIVE
  LOG_INFO(TELEMETRY_CATEGORY, "Resetting odometry");
  leftMotorEncoder.SetPosition(0);
  rightMotorEncoder.SetPosition(0);
  odometry.ResetPosition(frc::Pose2d(), imu->GetHeading());
#endif
}

const frc::Pose2d& PrivateDrive::GetPose() const {
  return odometry.GetPose();
}

bool PrivateDrive::ShiftHigh() {
  const auto state = GetCurrentState();
  if (state != SHIFTER_SHIFTING_LOW && state != SHIFTER_SHIFTING_HIGH) {
    ChangeState(SHIFTER_SHIFTING_HIGH);
    return true;
  }
  return false;
}

bool PrivateDrive::ShiftLow() {
  const auto state = GetCurrentState();
  if (state != SHIFTER_SHIFTING_LOW && state != SHIFTER_SHIFTING_HIGH) {
    ChangeState(SHIFTER_SHIFTING_LOW);
    return true;
  }
  return false;
}

bool PrivateDrive::IsHighGear() {
  return (GetCurrentState() == SHIFTER_HIGH_GEAR);
}

bool PrivateDrive::IsLowGear() {
  return (GetCurrentState() == SHIFTER_LOW_GEAR);
}

bool PrivateDrive::Claim(std::string_view systemName) {
  if (owner.length() == 0) {
    owner = systemName;
    LOG_INFO(TELEMETRY_CATEGORY, fmt::format("Now claimed by '{}'", systemName));
    return true;
  }
  else if (owner.compare(systemName) == 0) {
    LOG_WARNING(TELEMETRY_CATEGORY, fmt::format("'{}' tried to claim drive while already owned", systemName));
    return true;
  }

  LOG_WARNING(TELEMETRY_CATEGORY, fmt::format("'{}' tried to claim drive while owned by '{}'", systemName, owner));
  return false;
}

bool PrivateDrive::Release(std::string_view systemName) {
  if (owner.length() == 0) {
    LOG_WARNING(TELEMETRY_CATEGORY, fmt::format("Released by '{}' without owner", systemName));
    return true;
  }
  else if (owner.compare(systemName) == 0) {
    LOG_INFO(TELEMETRY_CATEGORY, fmt::format("Released by '{}'", systemName));
    owner = "";
    return true;
  }
  else {
    LOG_WARNING(TELEMETRY_CATEGORY, fmt::format("'{}' tried to release drive while owned by '{}'", systemName, owner));
    return false;
  }
}

void PrivateDrive::ArcadeDrive(double speed, double zRotation, bool squareInputs) {
#ifdef DOUBLE_SQUARE_TURNS
  zRotation = std::copysign(zRotation * zRotation, zRotation);
#endif
#ifdef ENABLE_DRIVE
  diffDrive.ArcadeDrive(-speed, zRotation, squareInputs);
#endif
}

void PrivateDrive::CurvatureDrive(double speed, double zRotation, bool quickTurn) {
#ifdef ENABLE_DRIVE
  diffDrive.CurvatureDrive(-speed, zRotation, quickTurn);
#endif
}

void PrivateDrive::DirectDrive(double leftMotorOutput, double rightMotorOutput) {
#ifdef ENABLE_DRIVE
  SetOutput(leftMotorOutput, rightMotorOutput);
#endif
}

void PrivateDrive::SetOutput(double leftMotorOutput, double rightMotorOutput) {
#ifdef ENABLE_DRIVE
  leftMotorLeader.Set(leftMotorOutput);
  rightMotorLeader.Set(rightMotorOutput);
#endif
}

void PrivateDrive::Stop() {
#ifdef ENABLE_DRIVE
  leftMotorLeader.StopMotor();
  rightMotorLeader.StopMotor();
#endif
}

Drive::Drive(std::string_view ownerName) : owner(ownerName) {
  LOG_INFO(TELEMETRY_CATEGORY, fmt::format("Automatically acquring for '{}'", ownerName));
  Get()->Claim(owner);
}

Drive::Drive(const Updateable* owner) : Drive(owner->GetName()) {}

Drive::~Drive() {
  LOG_INFO(TELEMETRY_CATEGORY, fmt::format("Automatically releasing from '{}'", owner));
  Get()->Release(owner);
}

void Drive::Init() {
  if (drive == nullptr) {
    drive = new PrivateDrive();
  }
  else {
    LOG_WARNING(TELEMETRY_CATEGORY, "Attempt to initialize Drive subsystem multiple times");
  }
}

PrivateDrive* Drive::Get() {
  if (drive == nullptr) {
    LOG_ERROR(TELEMETRY_CATEGORY, "Not initialized, will initialize inline (possible race condition)");
    drive = new PrivateDrive();
  }
  return drive;
}

PrivateDrive* Drive::operator->() {
  const auto d = Get();
  const auto sysOwner = d->GetOwner();
  if (sysOwner.compare(owner) != 0) {
    LOG_ERROR(TELEMETRY_CATEGORY, fmt::format("System '{}' attempting to control drive while '{}' is owner [!]", owner, sysOwner));
  }
  return d;
}

PrivateDrive* Drive::drive = nullptr;
