/**
 *  ____ _____ ___ ____            ____      _          _
 * |___ \___  / _ \___ \          |  _ \ ___| |__   ___| |___
 *   __) | / / | | |__) |  _____  | |_) / _ \ '_ \ / _ \ / __|
 *  / __/ / /| |_| / __/  |_____| |  _ <  __/ |_) |  __/ \__ \
 * |_____/_/  \___/_____|         |_| \_\___|_.__/ \___|_|___/
 *
 *
 * Copyright 2019 Kitchener-Waterloo Youth Robotics
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not
 * use this file except in compliance with the License. You may obtain a copy of
 * the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#pragma once
#include <frc/DoubleSolenoid.h>
#include <frc/drive/DifferentialDrive.h>
#include <frc/geometry/Pose2d.h>
#include <frc/kinematics/DifferentialDriveOdometry.h>
#include <frc/kinematics/DifferentialDriveKinematics.h>
#include <rev/CANSparkMax.h>

#include "Constants.h"
#include "Imu.h"
#include "MovingAverage.h"
#include "StateMachine.h"

namespace Rebels {

  enum ShifterState {
    SHIFTER_INIT,
    SHIFTER_LOW_GEAR,
    SHIFTER_SHIFTING_LOW,
    SHIFTER_HIGH_GEAR,
    SHIFTER_SHIFTING_HIGH,
  };

  const StateMachine<ShifterState>::StateStrings ShifterStrings = {
      "Init", "Low Gear", "Shifting Low", "High Gear", "Shifting High",
  };

  // PrivateDrive wraps drive functions that we want to ensure is owned by one
  // subsystem at a time. Two systems driving the motors at the same time makes
  // things go badly.
  class PrivateDrive : public StateMachine<ShifterState> {
    friend class Drive;

  public:
    void ArcadeDrive(double speed, double zRotation, bool squareInputs);
    void CurvatureDrive(double speed, double zRotation, bool quickTurn);
    void DirectDrive(double leftMotor, double rightMotor);

    void Stop();

    bool ShiftHigh();
    bool ShiftLow();
    bool IsHighGear();
    bool IsLowGear();

    bool IsCoastMode() const;
    bool IsBrakeMode() const;
    void SetBrakeMode();
    void SetCoastMode();

    void SetRotationTranslationFactor(double gearFactor);

    const frc::Pose2d& GetPose() const;
    void ResetOdometry();

    std::string_view GetOwner() const { return owner; }

  public:
    virtual void Update() override;
    virtual void BackgroundUpdate() override;

  protected:
    virtual bool TransitionTo(ShifterState newState) override;

  private:
    PrivateDrive();
    virtual ~PrivateDrive();

    // Claim the drive system, which will lock out other systems from using it
    bool Claim(std::string_view systemName);
    bool Release(std::string_view systemName);
    void SetOutput(double leftMotorOutput, double rightMotorOutput);
    void SetShifterSolenoid(frc::DoubleSolenoid::Value v);

    std::string_view owner;
    units::inch_t rotationTranslationFactor{ 1 };
    bool coastMode{ true };

#ifdef ENABLE_DRIVE
    // Brushless REV Robotics NEO SparkMax motors
    rev::CANSparkMax leftMotorLeader{ DRIVE_LEFT_LEADER_ID, rev::CANSparkMax::MotorType::kBrushless };
    rev::CANSparkMax rightMotorLeader{ DRIVE_RIGHT_LEADER_ID, rev::CANSparkMax::MotorType::kBrushless };
    rev::CANSparkMax leftMotorFollower{ DRIVE_LEFT_FOLLOWER_ID, rev::CANSparkMax::MotorType::kBrushless };
    rev::CANSparkMax rightMotorFollower{ DRIVE_RIGHT_FOLLOWER_ID, rev::CANSparkMax::MotorType::kBrushless };
    rev::SparkMaxRelativeEncoder leftMotorEncoder = leftMotorLeader.GetEncoder();
    rev::SparkMaxRelativeEncoder rightMotorEncoder = rightMotorLeader.GetEncoder();

#ifdef ENABLE_SHIFTERS
    // Gear Shifter Solenoid
    frc::DoubleSolenoid shifterSolenoid{ SHIFTER_MODULE_TYPE, SHIFTER_FORWARD_PCM_ID, SHIFTER_BACKWARD_PCM_ID };
#endif

    // Odometry, Kinematics, Differential drive
    frc::DifferentialDrive diffDrive{ leftMotorLeader, rightMotorLeader };
#endif
    frc::DifferentialDriveOdometry odometry{ frc::Rotation2d() };
    frc::DifferentialDriveKinematics kinematics{ DRIVE_WHEEL_BASE };

    std::unique_ptr<Rebels::Imu> imu;
  };

  // The Drive class wraps the main Drive code with an automatic acquire and
  // release. If you want to avoid log spam, make sure you only have one instance
  // of this created at any given time.
  class Drive {
  public:
    static void Init();

    Drive(std::string_view ownerName);
    Drive(const Updateable* owner);
    ~Drive();

    PrivateDrive* operator->();

  private:
    Drive() = delete;
    Drive(const Drive&) = delete;
    Drive(const Drive&&) = delete;
    Drive& operator=(const Drive&) = delete;
    Drive& operator=(const Drive&&) = delete;

    const std::string_view owner;
    static PrivateDrive* Get();
    static PrivateDrive* drive;
  };

} // namespace Rebels
