/**
 * Copyright 2019 Kitchener-Waterloo Youth Robotics
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not
 * use this file except in compliance with the License. You may obtain a copy of
 * the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "Imu.h"
#include "Settings.h"
#include "StateMachines.h"
#include "Telemetry.h"

using namespace Rebels;

#define TELEMETRY_CATEGORY "IMU"

Imu::Imu() : Updateable("Imu", HighUpdateablePriority) {
  Rebels::StateMachines::Register(this);
#ifdef ENABLE_IMU
  LOG_INFO(TELEMETRY_CATEGORY, "Calibrating the IMU");
  imu.EnterCalibrationMode(PigeonIMU::CalibrationMode::BootTareGyroAccel);
#endif
}

Imu::~Imu() {
  Rebels::StateMachines::Unregister(this);
}

void Imu::Update() {
  if (!imuStable) {
    CheckIMUReadiness();
  }
}

void Imu::BackgroundUpdate() {
#ifdef ENABLE_IMU
  PigeonIMU::GeneralStatus status;
  imu.GetGeneralStatus(status);

  std::string_view pigeonState{ "?" };
  switch (status.state) {
  case PigeonIMU::PigeonState::Initializing:
    pigeonState = "Init";
    break;
  case PigeonIMU::PigeonState::NoComm:
    pigeonState = "No comms";
    break;
  case PigeonIMU::PigeonState::Ready:
    pigeonState = "Ready";
    break;
  case PigeonIMU::PigeonState::UserCalibration:
    pigeonState = "Calibration";
    break;
  }

  LOG_VALUE(TELEMETRY_CATEGORY, "State", pigeonState);
  LOG_VALUE(TELEMETRY_CATEGORY, "Ready", pigeonState == "Ready");
  LOG_VALUE(TELEMETRY_CATEGORY, "Status", status.description);
  LOG_VALUE(TELEMETRY_CATEGORY, "Temperature", status.tempC);
  LOG_VALUE(TELEMETRY_CATEGORY, "Fused Heading", imu.GetFusedHeading());
#else
  LOG_VALUE(TELEMETRY_CATEGORY, "State", "Disabled");
#endif
  return;
}

frc::Rotation2d Imu::GetHeading() {
#ifdef ENABLE_IMU
  const double headingDegrees = imu.GetFusedHeading();
  return frc::Rotation2d(units::degree_t(headingDegrees));
#endif
  return frc::Rotation2d(units::degree_t(0));
}

double Imu::GetHeadingDegrees() {
#ifdef ENABLE_IMU
  return imu.GetFusedHeading();
#endif
  return 0;
}

double Imu::GetYaw() {
#ifdef ENABLE_IMU
  double ypr[3];
  imu.GetYawPitchRoll(ypr);
  return ypr[0];
#endif
  return 0;
}

double Imu::GetPitch() {
#ifdef ENABLE_IMU
  double ypr[3];
  imu.GetYawPitchRoll(ypr);
  return ypr[1];
#endif
  return 0;
}

double Imu::GetRoll() {
#ifdef ENABLE_IMU
  double ypr[3];
  imu.GetYawPitchRoll(ypr);
  return ypr[2];
#endif
  return 0;
}

bool Imu::IsReady() {
#ifdef ENABLE_IMU
  return imuStable;
#endif
  return true;
}

void Imu::CheckIMUReadiness() {
  if (imuStable) {
    return;
  }
#ifdef ENABLE_IMU
  PigeonIMU::GeneralStatus status;
  imu.GetGeneralStatus(status);
  if (status.state != PigeonIMU::PigeonState::Ready) {
    return;
  }

  // IMU is now ready -- wait for it to become stable
  const double h = imu.GetFusedHeading();
  if (std::fabs(h - lastImuReading) < 0.3) {
    const auto now = frc::Timer::GetFPGATimestamp();
    if (lastImuStableTime < 1_ms) {
      lastImuStableTime = now;
      LOG_INFO(TELEMETRY_CATEGORY, "The IMU has finished calibrating, waiting for it to become stable");
    }
    else if (now - lastImuStableTime > 3_s) {
      // IMU has been stable for 3s, we can consider it ready to use
      imuStable = true;
      LOG_INFO(TELEMETRY_CATEGORY, "The IMU has been stable for 3 seconds");
    }
  }
  else {
    lastImuReading = h;
    lastImuStableTime = 0_s;
  }
#endif
}
