/**
 * Copyright 2019 Kitchener-Waterloo Youth Robotics
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not
 * use this file except in compliance with the License. You may obtain a copy of
 * the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "Led.h"
#include "Telemetry.h"
#include "StateMachines.h"

#define TELEMETRY_CATEGORY "Led"

struct AroundTheWorld : public LedPattern {
  AroundTheWorld() : LedPattern("AroundTheWorld") {}
  virtual ~AroundTheWorld() = default;

  virtual bool Update(LedVector& data) {
    static const constexpr double dim = 0.4;
    static const int tail = 4;
    int prevPixels[tail];
    frc::Color prevColors[tail];
    prevPixels[0] = pixel;
    prevColors[0] = color;
    for (int i = 1; i < tail; ++i) {
      prevPixels[i] = prevPixels[i - 1] - direction;
      prevColors[i] =
        frc::Color(prevColors[i - 1].red * dim, prevColors[i - 1].green * dim,
          prevColors[i - 1].blue * dim);
    }

    for (int i = 0; i < length; i++) {
      data[i].SetRGB(0, 0, 0);
    }
    for (int j = 0; j < tail; ++j) {
      if (prevPixels[j] < 0 || prevPixels[j] >= length) {
        continue;
      }
      data[prevPixels[j]].SetLED(prevColors[j]);
    }
    pixel += direction;
    if (pixel >= length || pixel < 0) {
      direction = -1 * direction;
      pixel += direction;
      pixel += direction;
    }
    return true;
  }

  int pixel{ 0 };
  int direction{ 1 };
};

struct DanceParty : public LedPattern {
  DanceParty() : LedPattern("DanceParty") {}
  virtual ~DanceParty() = default;

  virtual bool Update(LedVector& data) {
    for (int i = 0; i < length; i++) {
      data[i].SetRGB(rand() % 200, rand() % 200, rand() % 200);
    }
    return true;
  }
};

struct Meter : public LedPattern {
  Meter() : LedPattern("Meter") {}
  virtual ~Meter() = default;

  virtual bool Update(LedVector& data) {
    const int pixels = percent * length;
    for (int i = 0; i < length; i++) {
      if (i < pixels) {
        data[i].SetLED(color);
      }
      else {
        data[i].SetLED(secondaryColor);
      }
    }
    return true;
  }

  double percent{ 0.0 };
  frc::Color secondaryColor;
};

struct Flasher : public LedPattern {
  Flasher() : LedPattern("Flasher") {}
  virtual ~Flasher() = default;

  virtual bool Update(LedVector& data) {
    for (int i = 0; i < length; i++) {
      if (on) {
        data[i].SetLED(color);
      }
      else {
        data[i].SetRGB(0, 0, 0);
      }
    }
    on = !on;
    count = 0;
    return true;
  }

  bool on{ false };
};

Led::Led(int pwmId, int length) : Updateable("Led", Rebels::LowUpdateablePriority), leds(pwmId), length(length) {
  LOG_INFO(TELEMETRY_CATEGORY, fmt::format("Creating LED strip on PWM {} with {} leds", pwmId, length));

  data.resize(length);
  for (int i = 0; i < length; ++i) {
    data[i].SetRGB(0, 0, 0);
  }
  leds.SetLength(length);
  leds.SetData(data);
  leds.Start();

  Rebels::StateMachines::Register(this);
}

Led::~Led() {
  Stop();
  Rebels::StateMachines::Unregister(this);
}

void Led::Update() {
  std::lock_guard<std::mutex> guard(patternMutex);
  if (pattern != nullptr) {
    if (++pattern->count <= pattern->delayCount) {
      return;
    }
    pattern->count = 0;
    if (pattern->Update(data)) {
      leds.SetData(data);
    }
  }
}

void Led::StartAroundTheWorld(frc::Color color, units::time::millisecond_t delay) {
  Stop();
  std::lock_guard<std::mutex> guard(patternMutex);
  pattern = new AroundTheWorld();
  pattern->color = color;
  pattern->length = length;
  pattern->delayCount = std::clamp(delay.to<int>() / 20, 1, 100);
}

void Led::StartFlasher(frc::Color color, units::time::millisecond_t delay) {
  Stop();
  std::lock_guard<std::mutex> guard(patternMutex);
  pattern = new Flasher();
  pattern->color = color;
  pattern->length = length;
  pattern->delayCount = std::clamp(delay.to<int>() / 20, 1, 100);
}

void Led::StartDanceParty(units::time::millisecond_t delay) {
  Stop();
  std::lock_guard<std::mutex> guard(patternMutex);
  pattern = new DanceParty();
  pattern->length = length;
  pattern->delayCount = std::clamp(delay.to<int>() / 20, 1, 100);
}

void Led::SetMeter(frc::Color mainColor, frc::Color secondaryColor, double percent) {
  std::lock_guard<std::mutex> guard(patternMutex);
  if (dynamic_cast<Meter*>(pattern) == nullptr) {
    if (pattern != nullptr) {
      delete pattern;
    }
    pattern = new Meter();
    pattern->length = length;
  }
  pattern->color = mainColor;
  (dynamic_cast<Meter*>(pattern))->percent = percent;
  (dynamic_cast<Meter*>(pattern))->secondaryColor = secondaryColor;
}

void Led::Stop() {
  std::lock_guard<std::mutex> guard(patternMutex);
  if (pattern != nullptr) {
    delete pattern;
  }
  pattern = nullptr;
}
