/**
 * Copyright 2019 Kitchener-Waterloo Youth Robotics
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not
 * use this file except in compliance with the License. You may obtain a copy of
 * the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <frc/DriverStation.h>
#include <frc/Timer.h>

#include "Log.h"

using namespace Rebels;

StaticLogger::StaticLogger() {
#if defined(ENABLE_TIMED_LOGS) || defined(ENABLE_DEBUG_LOGS)
  queueThread = std::thread(&StaticLogger::writeQueueThread, this);
#endif
}

void StaticLogger::Info(const char* fmt, ...) {
  va_list args;
  va_start(args, fmt);
  Get()->log(LogLevelInfo, fmt, args);
  va_end(args);
}

void StaticLogger::Warning(const char* fmt, ...) {
  va_list args;
  va_start(args, fmt);
  Get()->log(LogLevelWarning, fmt, args);
  va_end(args);
}

void StaticLogger::Error(const char* fmt, ...) {
  va_list args;
  va_start(args, fmt);
  Get()->log(LogLevelError, fmt, args);
  va_end(args);
}

void StaticLogger::Debug(const char* fmt, ...) {
  va_list args;
  va_start(args, fmt);
  Get()->log(LogLevelDebug, fmt, args);
  va_end(args);
}

void StaticLogger::log(Level level, const char* fmt, va_list args) {
  char buffer[1024];
  const int w = vsnprintf(buffer, sizeof(buffer) - 1, fmt, args);
  if (w < 0) {
    printf("ERROR: Error parsing buffer for Log()\n");
    return;
  }
  buffer[sizeof(buffer) - 1] = '\0';
  log2(level, buffer);
}

void StaticLogger::log2(Level level, const char* buf) {
  char buffer[1024];
  const size_t len = strlen(buf);
  sprintf(buffer, "%8.03f: %s%s%s",
    frc::Timer::GetFPGATimestamp().to<double>(),
    (level == LogLevelWarning
      ? warningStr
      : (level == LogLevelDebug
        ? debugStr
        : (level == LogLevelDebug ? debugStr : ""))),
    buf, ((len == 0 || buf[len - 1] != '\n') ? "\n" : ""));

  if (level != LogLevelDebug) {
    printf("%s", buffer);
  }

#if defined(ENABLE_TIMED_LOGS) || defined(ENABLE_DEBUG_LOGS)
  std::lock_guard<std::mutex> lock(queueMutex);
  queue.push_back(buffer);
#endif
}

#if defined(ENABLE_TIMED_LOGS) || defined(ENABLE_DEBUG_LOGS)
void StaticLogger::writeQueueThread() {
  frc::Wait(5.0_s);
  const auto waitTime = 0.1_s;

  while (true) {
    const int matchNumber = frc::DriverStation::GetMatchNumber();

    char filename[256];
#if defined(ENABLE_TIMED_LOGS)
    const std::time_t t = std::time(0);
    const std::tm* now = std::localtime(&t);
    sprintf(filename, "/home/lvuser/frc_%d_%04d-%02d-%02d_%02d%02d%02d.log",
      matchNumber, now->tm_year + 1900, now->tm_mon + 1, now->tm_mday,
      now->tm_hour, now->tm_min, now->tm_sec);
    FILE* f = fopen(filename, "w+");
    if (!f) {
      printf("[ERROR] Failed to open debug log '%s'!\n", filename);
      return;
    }
#endif

#if defined(ENABLE_DEBUG_LOGS)
    sprintf(filename, "/home/lvuser/frc_debug.log");
    FILE* fDebug = fopen(filename, "a+");
    if (!fDebug) {
      printf("[ERROR] Failed to open debug log '%s'!\n", filename);
      return;
    }
#endif
    std::vector<std::string> local;

    while (matchNumber == frc::DriverStation::GetMatchNumber()) {
      pthread_yield();
      frc::Wait(waitTime);
      if (true) {
        std::lock_guard<std::mutex> lock(queueMutex);
        local = queue;
        queue.clear();
      }
      if (local.empty()) {
        continue;
      }
      for (const auto& str : local) {
#if defined(ENABLE_TIMED_LOGS)
        fprintf(f, "%s", str.c_str());
#endif
#if defined(ENABLE_DEBUG_LOGS)
        fprintf(fDebug, "%s", str.c_str());
#endif
      }
#if defined(ENABLE_TIMED_LOGS)
      fflush(f);
#endif
#if defined(ENABLE_DEBUG_LOGS)
      fflush(fDebug);
#endif
    }
#if defined(ENABLE_TIMED_LOGS)
    fclose(f);
#endif
#if defined(ENABLE_DEBUG_LOGS)
    fclose(fDebug);
#endif
  }
}
#endif
