/**
 *  ____ _____ ___ ____            ____      _          _
 * |___ \___  / _ \___ \          |  _ \ ___| |__   ___| |___
 *   __) | / / | | |__) |  _____  | |_) / _ \ '_ \ / _ \ / __|
 *  / __/ / /| |_| / __/  |_____| |  _ <  __/ |_) |  __/ \__ \
 * |_____/_/  \___/_____|         |_| \_\___|_.__/ \___|_|___/
 *
 *
 * Copyright 2019 Kitchener-Waterloo Youth Robotics
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not
 * use this file except in compliance with the License. You may obtain a copy of
 * the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#pragma once

#include <stdarg.h>
#include <mutex>
#include <string>
#include <thread>
#include <vector>

#define ENABLE_TIMED_LOGS
#define ENABLE_DEBUG_LOGS

#include "Singleton.h"

namespace Rebels {

  class StaticLogger {
    DECLARE_SINGLETON(StaticLogger)

  public:
    enum Level {
      LogLevelInfo,
      LogLevelWarning,
      LogLevelError,
      LogLevelDebug,
    };
    StaticLogger();
    ~StaticLogger() = default;

    static void Info(const char* fmt, ...) __attribute__((format(printf, 1, 2)));
    static void Warning(const char* fmt, ...) __attribute__((format(printf, 1, 2)));
    static void Error(const char* fmt, ...) __attribute__((format(printf, 1, 2)));
    static void Debug(const char* fmt, ...) __attribute__((format(printf, 1, 2)));

  private:
    void log(Level level, const char* fmt, va_list args);
    void log2(Level level, const char* str);

  private:
    const char* warningStr = "[WARNING] ";
    const char* errorStr = "[ERROR] ";
    const char* debugStr = "[DEBUG] ";
#if defined(ENABLE_TIMED_LOGS) || defined(ENABLE_DEBUG_LOGS)
    void writeQueueThread();
    std::vector<std::string> queue;
    std::mutex queueMutex;
    std::thread queueThread;
#endif
  };

} // namespace Rebels

#define LOG_INFO Rebels::StaticLogger::Info
#define LOG_WARNING Rebels::StaticLogger::Warning
#define LOG_ERROR Rebels::StaticLogger::Error
#define LOG_DEBUG Rebels::StaticLogger::Debug
