/**
 * Copyright 2019 Kitchener-Waterloo Youth Robotics
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not
 * use this file except in compliance with the License. You may obtain a copy of
 * the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <arpa/inet.h>
#include <netinet/in.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <unistd.h>
#include <frc/DriverStation.h>
#include <frc/Timer.h>

#include "PiBroadcaster.h"
#include "Settings.h"
#include "StateMachines.h"
#include "Telemetry.h"

using namespace Rebels;

#define TELEMETRY_CATEGORY "Pi Broadcaster"

REGISTER_STRING(CommandString, "Vision: Pi Broadcast Commands", "HC")

struct Socket {
  Socket() {
    fd = socket(AF_INET, SOCK_DGRAM, 0);
  }

  ~Socket() {
    if (fd >= 0) {
      close(fd);
      fd = -1;
    }
  }

  int fd{ -1 };
};

PiBroadcaster::PiBroadcaster() : Updateable("PiBroadcaster") {
  Rebels::StateMachines::Register(this);
}

void PiBroadcaster::BackgroundUpdate() {
  const auto now = frc::Timer::GetFPGATimestamp();
  if (now - lastBroadcast < broadcastPeriod) {
    return;
  }
  lastBroadcast = now;

  struct sockaddr_in send_addr;
  int trueflag = 1;
  Socket s;
  if (s.fd < 0) {
    LOG_INFO(TELEMETRY_CATEGORY, fmt::format("Failed to create socket: {}", strerror(errno)));
    return;
  }

  if (setsockopt(s.fd, SOL_SOCKET, SO_REUSEADDR, &trueflag, sizeof trueflag) < 0) {
    LOG_INFO(TELEMETRY_CATEGORY, fmt::format("Failed to set socket options to reuse addresses: {}", strerror(errno)));
    return;
  }

  if (setsockopt(s.fd, SOL_SOCKET, SO_BROADCAST, &trueflag, sizeof trueflag) < 0) {
    LOG_INFO(TELEMETRY_CATEGORY, fmt::format("Failed to set socket options to broadcast: {}", strerror(errno)));
    return;
  }

  memset(&send_addr, 0, sizeof send_addr);
  send_addr.sin_family = AF_INET;
  send_addr.sin_port = (in_port_t)htons(2702);

  // broadcasting address for unix (?)
  inet_aton("10.27.2.255", &send_addr.sin_addr);

  // send_addr.sin_addr.s_addr = htonl(INADDR_BROADCAST);
  const auto buf = fmt::format("{}M{:d}", GetCommandString(), frc::DriverStation::GetMatchNumber());
  const auto res = sendto(s.fd, buf.c_str(), buf.length() + 1, 0, (struct sockaddr*)&send_addr, sizeof send_addr);
  if (res < 0) {
    LOG_INFO(TELEMETRY_CATEGORY, fmt::format("Failed to send: {}", strerror(errno)));
  }
  else {
    LOG_DEBUG(TELEMETRY_CATEGORY, fmt::format("Sent command string '{}'", buf));
  }
}
