#include <frc/Preferences.h>
#include <frc/SmartDashboard/SmartDashboard.h>
#include <frc/Timer.h>
#include <sys/stat.h>
#include <sys/types.h>

#include "Telemetry.h"
#include "Settings.h"
#include "StateMachines.h"

#define TELEMETRY_CATEGORY "Settings"

static const constexpr std::string_view kEnableWriteDash{ "Allow Settings Changes" };

Setting::Setting() {
  val.strValue = nullptr;
  type = SETTING_NONE;
}

Setting::~Setting() {
  if (type == SETTING_STRING && val.strValue != nullptr) {
    free(val.strValue);
  }
}

Setting::Setting(const std::string& str) { SetString(str.c_str()); }

Setting::Setting(const double d) { SetDouble(d); }

Setting::Setting(const bool b) { SetBool(b); }

Setting::Setting(const Setting& that) { Copy(that); }

Setting& Setting::operator=(const Setting& that) {
  Copy(that);
  return *this;
}

void Setting::Copy(const Setting& that) {
  switch (that.type) {
  case SETTING_NONE:
    break;
  case SETTING_STRING:
    SetString(that.val.strValue);
    break;
  case SETTING_DOUBLE:
    SetDouble(that.val.doubleValue);
    break;
  case SETTING_BOOL:
    SetBool(that.val.boolValue);
    break;
  }
}

bool Setting::IsString() const {
  return (type == SETTING_STRING && val.strValue != nullptr);
}

bool Setting::IsDouble() const {
  return (type == SETTING_DOUBLE);
}

bool Setting::IsBool() const {
  return (type == SETTING_BOOL);
}

std::string Setting::GetString() const {
  if (IsString()) {
    return val.strValue;
  }

  LOG_ERROR(TELEMETRY_CATEGORY, "GetString() called on empty/non-string value");
  return "";
}

double Setting::GetDouble() const {
  if (IsDouble()) {
    return val.doubleValue;
  }

  LOG_ERROR(TELEMETRY_CATEGORY, "GetDouble() called on non-double value");
  return 0.0f;
}

bool Setting::GetBool() const {
  if (IsBool()) {
    return val.boolValue;
  }

  LOG_ERROR(TELEMETRY_CATEGORY, "GetBool() called on non-bool value");
  return false;
}

void Setting::SetString(const char* str) {
  if (type == SETTING_STRING && val.strValue != nullptr) {
    free(val.strValue);
    val.strValue = nullptr;
  }

  type = SETTING_STRING;
  if (str != nullptr) {
    val.strValue = strdup(str);
  }
}

void Setting::SetDouble(const double d) {
  type = SETTING_DOUBLE;
  val.doubleValue = d;
}

void Setting::SetBool(const bool b) {
  type = SETTING_BOOL;
  val.boolValue = b;
}

Settings::Settings() : Updateable("Settings") {
  Init();
  Rebels::StateMachines::Register(this);
}

void Settings::BackgroundUpdate() {
  Get()->UpdateTable();
}

void Settings::DisableWrites() {
  frc::SmartDashboard::PutBoolean(kEnableWriteDash, false);
  Get()->changesAreAllowed = true;
  LOG_INFO(TELEMETRY_CATEGORY, "Writes are disabled");
}

void Settings::EnableWrites() {
  Get()->changesAreAllowed = true;
  LOG_INFO(TELEMETRY_CATEGORY, "Writes are enabled");
}

void Settings::Init() {
  std::lock_guard<std::mutex> lock(initMutex);
  if (isInitialized) {
    return;
  }

  LOG_INFO(TELEMETRY_CATEGORY, "Initializing");

  const auto res = mkdir(SETTINGS_PATH, S_IRWXU | S_IRWXU);
  if (res != 0 && errno != EEXIST) {
    LOG_ERROR(TELEMETRY_CATEGORY, fmt::format("Failed to create folder '{}'", SETTINGS_PATH));
  }

  frc::SmartDashboard::PutBoolean(kEnableWriteDash, false);
  changesAreAllowed = false;
  isInitialized = true;
}

void Settings::UpdateTable() {
  // If changes are disallowed, just update that state and return
  // Also turn off a flag that was left on.
  if (!changesAreAllowed || frc::Timer::GetFPGATimestamp() < 2_s) {
    if (frc::SmartDashboard::GetBoolean(kEnableWriteDash, false)) {
      LOG_INFO(TELEMETRY_CATEGORY, fmt::format("Clearing '{}'", kEnableWriteDash));
      frc::SmartDashboard::PutBoolean(kEnableWriteDash, false);
    }
    openForChanges = false;
    return;
  }

  const bool open = frc::SmartDashboard::GetBoolean(kEnableWriteDash, false);
  if (open && openForChanges && lastSettingsChange > 10_ms &&
    frc::Timer::GetFPGATimestamp() - lastSettingsChange > 60_s) {
    LOG_INFO(TELEMETRY_CATEGORY, "No settings have been changed in 60 seconds, disabling writes");
    frc::SmartDashboard::PutBoolean(kEnableWriteDash, false);
    openForChanges = false;
    return;
  }

  if (open && !openForChanges) {
    LOG_WARNING(TELEMETRY_CATEGORY, "System is now open for changes!");
    lastSettingsChange = frc::Timer::GetFPGATimestamp();
  }
  else if (!open && openForChanges) {
    LOG_WARNING(TELEMETRY_CATEGORY, "System is now closed for changes");
  }

  openForChanges = open;

  // Remove any preferences that aren't ours.
  {
    std::lock_guard<std::mutex> lock(kvMapMutex);
    auto keys = frc::Preferences::GetKeys();
    for (auto key : keys) {
      if (key.compare(kEnableWriteDash) == 0) {
        continue;
      }
      auto it = kvMap.find(key);
      if (it == kvMap.end()) {
        if (key.compare(".type") != 0) {
          LOG_WARNING(TELEMETRY_CATEGORY, fmt::format("Removing unknown setting '{}'", key));
        }
        frc::Preferences::Remove(key);
      }
    }
  }

  // Update the preferences depending on if we're accepting changes or not.
  for (const auto& it : kvMap) {
    const auto& key = it.first;
    const auto& val = it.second;
    if (val.IsString()) {
      const auto got = frc::Preferences::GetString(key, "");
      const auto ours = GetString(key);
      if (ours.compare(got) != 0) {
        lastSettingsChange = frc::Timer::GetFPGATimestamp();
        if (open) {
          LOG_WARNING(TELEMETRY_CATEGORY, fmt::format("Updating string '{}' = '{}'", key, got));
          SetString(key, got);
        }
        else {
          LOG_WARNING(TELEMETRY_CATEGORY, fmt::format("Forcing string '{}' = '{}'", key, ours));
          frc::Preferences::SetString(key, ours);
        }
      }
    }
    else if (val.IsDouble()) {
      const auto got = frc::Preferences::GetDouble(key, 0);
      const auto ours = GetDouble(key);
      if (got != ours) {
        lastSettingsChange = frc::Timer::GetFPGATimestamp();
        if (open) {
          LOG_WARNING(TELEMETRY_CATEGORY, fmt::format("Updating double '{}' = {:f}", key, got));
          SetDouble(key, got);
        }
        else {
          LOG_WARNING(TELEMETRY_CATEGORY, fmt::format("Forcing double '{}' = {:f}", key, ours));
          frc::Preferences::SetDouble(key, ours);
        }
      }
    }
    else if (val.IsBool()) {
      const auto got = frc::Preferences::GetBoolean(key, false);
      const auto ours = GetBool(key);
      if (got != ours) {
        lastSettingsChange = frc::Timer::GetFPGATimestamp();
        if (open) {
          LOG_WARNING(TELEMETRY_CATEGORY, fmt::format("Updating bool '{}' = {}", key, got));
          SetBool(key, got);
        }
        else {
          LOG_WARNING(TELEMETRY_CATEGORY, fmt::format("Forcing bool '{}' = {}", key, ours));
          frc::Preferences::SetBoolean(key, ours);
        }
      }
    }
  }
}

std::string Settings::GetFilePath(std::string_view key) const {
  std::string s{ key };
  for (unsigned i = 0; i < s.length(); ++i) {
    if (!std::isalnum(static_cast<unsigned char>(s[i]))) {
      s[i] = '_';
    }
  }

  return fmt::format("{}/{}.cfg", SETTINGS_PATH, s);
}

void Settings::SetMapAndPrefsString(std::string_view key, const std::string& value) {
  std::lock_guard<std::mutex> lock(kvMapMutex);
  kvMap[key] = Setting(value);
  frc::Preferences::SetString(key, value);
}

void Settings::SetMapAndPrefsDouble(std::string_view key, const double value) {
  std::lock_guard<std::mutex> lock(kvMapMutex);
  kvMap[key] = Setting(value);
  frc::Preferences::SetDouble(key, value);
}

void Settings::SetMapAndPrefsBool(std::string_view key, const bool value) {
  std::lock_guard<std::mutex> lock(kvMapMutex);
  kvMap[key] = Setting(value);
  frc::Preferences::SetBoolean(key, value);
}

void Settings::RegisterString(std::string_view key, const std::string& defaultValue) {
  Init();
  LOG_DEBUG(TELEMETRY_CATEGORY, fmt::format("Registered string '{}' with default '{}'", key, defaultValue));

  std::string value;
  if (Load(key, value)) {
    SetMapAndPrefsString(key, value);
    return;
  }

  if (Save(key, defaultValue)) {
    SetMapAndPrefsString(key, defaultValue);
    return;
  }

  LOG_ERROR(TELEMETRY_CATEGORY, fmt::format("Failed to save key '{}'", key));
}

void Settings::RegisterDouble(std::string_view key, const double defaultValue) {
  Init();
  LOG_DEBUG(TELEMETRY_CATEGORY, fmt::format("Registered double '{}' with default '{}'", key, defaultValue));

  std::string valueStr;
  if (Load(key, valueStr)) {
    const double v = atof(valueStr.c_str());
    SetMapAndPrefsDouble(key, v);
    return;
  }

  const std::string defaultValueStr = DoubleToString(defaultValue);
  if (Save(key, defaultValueStr)) {
    SetMapAndPrefsDouble(key, defaultValue);
    return;
  }

  LOG_ERROR(TELEMETRY_CATEGORY, fmt::format("Failed to save key '{}'", key));
}

void Settings::RegisterBool(std::string_view key, const bool defaultValue) {
  Init();
  LOG_DEBUG(TELEMETRY_CATEGORY, fmt::format("Registered bool '{}' with default '{}'", key, defaultValue));

  std::string valueStr;
  if (Load(key, valueStr)) {
    const bool b = (valueStr.length() && valueStr[0] == '1');
    SetMapAndPrefsBool(key, b);
    return;
  }

  const std::string defaultValueStr = (defaultValue ? "1" : "0");
  if (Save(key, defaultValueStr)) {
    SetMapAndPrefsBool(key, defaultValue);
    return;
  }

  LOG_ERROR(TELEMETRY_CATEGORY, fmt::format("Failed to save key '{}'", key));
}

bool Settings::Save(std::string_view key, const std::string& value) {
  Init();
  const auto p = GetFilePath(key);
  const auto f = fopen(p.c_str(), "w+");
  if (f == nullptr) {
    return false;
  }

  fwrite(value.c_str(), value.length(), 1, f);
  fflush(f);
  fclose(f);
  return true;
}

bool Settings::Load(std::string_view key, std::string& value) {
  Init();
  const auto p = GetFilePath(key);
  const auto f = fopen(p.c_str(), "r");
  if (f == nullptr) {
    return false;
  }

  char buf[1024];
  const auto l = fread(buf, 1, sizeof(buf) - 1, f);
  buf[l] = 0;               // Paranoid
  buf[sizeof(buf) - 1] = 0; // Paranoid
  fclose(f);

  value = buf;
  return true;
}

bool Settings::SetString(std::string_view key, const std::string& value) {
  if (Save(key, value)) {
    std::lock_guard<std::mutex> lock(kvMapMutex);
    kvMap[key] = Setting(value);
    return true;
  }
  return false;
}

bool Settings::SetDouble(std::string_view key, const double value) {
  const std::string v = DoubleToString(value);
  if (Save(key, v)) {
    std::lock_guard<std::mutex> lock(kvMapMutex);
    kvMap[key] = Setting(value);
    return true;
  }
  return false;
}

bool Settings::SetBool(std::string_view key, const bool value) {
  const std::string v = value ? "1" : "0";
  if (Save(key, v)) {
    std::lock_guard<std::mutex> lock(kvMapMutex);
    kvMap[key] = Setting(value);
    return true;
  }
  return false;
}

std::string Settings::GetString(std::string_view key) const {
  std::lock_guard<std::mutex> lock(kvMapMutex);
  auto it = kvMap.find(key);
  if (it == kvMap.end()) {
    LOG_WARNING(TELEMETRY_CATEGORY, fmt::format("Failed GetString({})", key));
    return "";
  }
  return it->second.GetString();
}

double Settings::GetDouble(std::string_view key) const {
  std::lock_guard<std::mutex> lock(kvMapMutex);
  auto it = kvMap.find(key);
  if (it == kvMap.end()) {
    LOG_WARNING(TELEMETRY_CATEGORY, fmt::format("Failed GetDouble({})", key));
    return 0.0f;
  }
  return it->second.GetDouble();
}

bool Settings::GetBool(std::string_view key) const {
  std::lock_guard<std::mutex> lock(kvMapMutex);
  auto it = kvMap.find(key);
  if (it == kvMap.end()) {
    LOG_WARNING(TELEMETRY_CATEGORY, fmt::format("Failed GetBool({})", key));
    return false;
  }
  return it->second.GetBool();
}

std::string Settings::IntToString(int i) const {
  return fmt::format("{:d}", i);
}

std::string Settings::DoubleToString(double d) const {
  return fmt::format("{:f}", d);
}
