/**
 *  ____ _____ ___ ____            ____      _          _
 * |___ \___  / _ \___ \          |  _ \ ___| |__   ___| |___
 *   __) | / / | | |__) |  _____  | |_) / _ \ '_ \ / _ \ / __|
 *  / __/ / /| |_| / __/  |_____| |  _ <  __/ |_) |  __/ \__ \
 * |_____/_/  \___/_____|         |_| \_\___|_.__/ \___|_|___/
 *
 *
 * Copyright 2019 Kitchener-Waterloo Youth Robotics
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not
 * use this file except in compliance with the License. You may obtain a copy of
 * the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 *
 * A class to manage robot settings.
 *
 * This class extends the functionality of frc::Preferences in several ways:
 *  - Adds a dashboard boolean that must be checked in order for settings to be
 * changed. This prevents accidental settings wipes when another computer
 * connects that has older settings.
 *  - Stores the settings on the robot in an easy to find location. Suitable
 * for changing via SCP if desired.
 *  - Macros are provided for easy creation of settings options. These can
 * either be done at a file level where they will be created statically, or
 * inside a class. eg. REGISTER_BOOL(MySettingName, "My Setting", false);
 */

#pragma once

#include <map>
#include <mutex>
#include <string>

#include "Singleton.h"
#include "Updateable.h"

class Setting {
public:
  Setting();
  ~Setting();
  explicit Setting(const std::string& str);
  explicit Setting(const char* str);
  explicit Setting(const double d);
  explicit Setting(const bool b);

  Setting(const Setting&);
  Setting& operator=(const Setting&);

  bool IsString() const;
  bool IsDouble() const;
  bool IsBool() const;

  std::string GetString() const;
  double GetDouble() const;
  bool GetBool() const;

  void SetString(const char* str);
  void SetDouble(const double d);
  void SetBool(const bool b);

private:
  void Copy(const Setting&);

  union {
    char* strValue;
    double doubleValue;
    bool boolValue;
  } val;

  enum { SETTING_NONE, SETTING_STRING, SETTING_DOUBLE, SETTING_BOOL } type;
};

class Settings : public Rebels::Updateable {
  DECLARE_SINGLETON(Settings)

public:
  // Toggle the ability to write to settings. You probably want to only
  // allow writes when the robot is disabled.
  static void DisableWrites();
  static void EnableWrites();

  // Register a setting. This sets the default, or loads from disk if
  // available. Call this before SetXXX or GetXXX.
  // Prefer the macros to calling this directly (eg. REGISTER_DOUBLE)
  void RegisterString(std::string_view key, const std::string& defaultValue);
  void RegisterDouble(std::string_view key, const double defaultValue);
  void RegisterBool(std::string_view key, const bool defaultValue);

  // Set a value. You must use Register() beforehand, otherwise this will fail.
  bool SetString(std::string_view key, const std::string& value);
  bool SetDouble(std::string_view key, const double value);
  bool SetBool(std::string_view key, const bool value);

  // Get a value by key.
  std::string GetString(std::string_view key) const;
  double GetDouble(std::string_view key) const;
  bool GetBool(std::string_view key) const;

protected:
  Settings();
  ~Settings() = default;

  void Init();
  void UpdateTable();

  // Updateable
  virtual void Update() override {}
  virtual void BackgroundUpdate() override;

private:
  std::string GetFilePath(std::string_view key) const;
  std::string IntToString(int i) const;
  std::string DoubleToString(double d) const;

  void SetMapAndPrefsString(std::string_view key, const std::string& value);
  void SetMapAndPrefsDouble(std::string_view key, const double value);
  void SetMapAndPrefsBool(std::string_view key, const bool value);

  bool Save(std::string_view key, const std::string& value);
  bool Load(std::string_view key, std::string& value);

private:
  static const constexpr char* SETTINGS_PATH = "/home/lvuser/robotsettings";

  // Our internal map of values, which is the record of truth.
  std::map<std::string_view, Setting> kvMap;
  mutable std::mutex kvMapMutex;

  // If writes are currently enabled by the user.
  bool openForChanges = false;

  // If writing is allowed at all.
  bool changesAreAllowed = false;

  // Time of the last settings change.
  units::second_t lastSettingsChange{ 0 };

  bool isInitialized = false;
  std::mutex initMutex;
};

namespace Rebels {
  class RegisterBool {
  public:
    RegisterBool(const char* name, bool def) {
      Settings::Get()->RegisterBool(name, def);
    }
  };
} // namespace Rebels

#define DECLARE_BOOL(fn) bool Get##fn();

#define REGISTER_BOOL(fn, param, defValue)                   \
  bool Get##fn() { return Settings::Get()->GetBool(param); } \
  void Set##fn(bool v) { Settings::Get()->SetBool(param, v); } \
  static Rebels::RegisterBool obj##fn{param, defValue};

namespace Rebels {
  class RegisterDouble {
  public:
    RegisterDouble(const char* name, double def) {
      Settings::Get()->RegisterDouble(name, def);
    }
  };
}  // namespace Rebels

#define DECLARE_DOUBLE(fn) double Get##fn();

#define REGISTER_DOUBLE(fn, param, defValue)                       \
  double Get##fn() { return Settings::Get()->GetDouble(param); }   \
  void Set##fn(double v) { Settings::Get()->SetDouble(param, v); } \
  static Rebels::RegisterDouble obj##fn{param, defValue};

#define DECLARE_INT(fn) int Get##fn();

#define REGISTER_INT(fn, param, defValue)                                 \
  static int Get##fn() { return (int)Settings::Get()->GetDouble(param); } \
  void Set##fn(int v) { Settings::Get()->SetDouble(param, v); }           \
  static Rebels::RegisterDouble obj##fn{param, defValue};

namespace Rebels {
  class RegisterString {
  public:
    RegisterString(const char* name, const char* def) {
      Settings::Get()->RegisterString(name, def);
    }
  };
}  // namespace Rebels

#define DECLARE_STRING(fn) std::string Get##fn();

#define REGISTER_STRING(fn, param, defValue)                                 \
  static std::string Get##fn() { return Settings::Get()->GetString(param); } \
  void Set##fn(std::string v) { Settings::Get()->SetString(param, v); }      \
  static Rebels::RegisterString obj##fn{param, defValue};
