/**
 *  ____ _____ ___ ____            ____      _          _
 * |___ \___  / _ \___ \          |  _ \ ___| |__   ___| |___
 *   __) | / / | | |__) |  _____  | |_) / _ \ '_ \ / _ \ / __|
 *  / __/ / /| |_| / __/  |_____| |  _ <  __/ |_) |  __/ \__ \
 * |_____/_/  \___/_____|         |_| \_\___|_.__/ \___|_|___/
 *
 *
 * Copyright 2020 Kitchener-Waterloo Youth Robotics
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not
 * use this file except in compliance with the License. You may obtain a copy of
 * the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#pragma once

#include <frc/Timer.h>
#include <list>
#include <string>

#include "Telemetry.h"
#include "StateMachines.h"

#define TELEMETRY_CATEGORY_STATEMACHINE "State Machine"

namespace Rebels {

  template <class T>
  struct Path {
    T fromState, toState;
  };

  template <class T>
  struct StateTransition {
    T newState;
    bool forceTransition;
  };

  template <class T>
  class StateMachine : public Updateable {
  public:
    typedef std::vector<std::string_view> StateStrings;

    StateMachine(std::string_view name, const StateStrings& strings, UpdateablePriority priority = DefaultUpdateablePriority) : Updateable(name, priority),
      currentState(T(0)),
      defaultState(T(0)),
      stateStrings(strings) {
      lastChangeTimer.Start();
      lastChangeTimer.Reset();
      LOG_INFO(TELEMETRY_CATEGORY_STATEMACHINE, fmt::format("Created '{}' ({})", name, strings[currentState]));
      StateMachines::Register(this);
    }

    virtual ~StateMachine() {
      StateMachines::Unregister(this);
      LOG_INFO(TELEMETRY_CATEGORY_STATEMACHINE, fmt::format("Removed '{}'", name));
    }

    StateMachine(const StateMachine&) = delete;
    StateMachine& operator=(const StateMachine&) = delete;

    T GetCurrentState() const {
      return currentState;
    }

  protected:
    // Update is called frequently and should be handled by state machines.
    // This is a good time to update motors, transition to other states, etc.
    virtual void Update() override = 0;

    // TransitionTo is called when the state machine should transition to the new
    // state. The state machine has already verified that the new state is
    // acceptable. Returns true on success, false on failure.
    virtual bool TransitionTo(T newState) = 0;

    bool CanTransition(T fromState, T toState) const {
      for (auto p : allowed) {
        if (p.fromState == fromState && p.toState == toState) {
          return true;
        }
      }
      return false;
    }

    bool AddTransition(T fromState, T toState) {
      return AddTransition(Path<T>{fromState, toState});
    }

    bool AddTransition(Path<T> path) {
      for (auto p : allowed) {
        if (p.fromState == path.fromState && p.toState == path.toState) {
          return false;
        }
      }
      allowed.push_back(path);
      return true;
    }

    void ClearTransitions() {
      allowed.clear();
    }

    void Reset() {
      currentState = defaultState;
      TransitionTo(currentState);
      lastChangeTimer.Reset();
      LOG_VALUE(TELEMETRY_CATEGORY_STATEMACHINE, name, stateStrings[currentState]);
    }

    void ChangeState(const T newState) {
      SetState(StateTransition<T>{newState, false});
    }

    void ForceState(const T newState) {
      SetState(StateTransition<T>{newState, true});
    }

  protected:
    inline units::second_t ElapsedSinceTransition() const {
      return lastChangeTimer.Get();
    }

  private:
    void SetState(StateTransition<T> newStateTransition) {
      const auto cs = stateStrings[currentState];
      const auto rs = stateStrings[newStateTransition.newState];

      LOG_DEBUG(TELEMETRY_CATEGORY_STATEMACHINE, fmt::format("Request [{}]: {} -> {}", name, cs, rs));

      {
        std::lock_guard<std::mutex> lock(queueMutex);
        queue.push_back(newStateTransition);

        // If this is a recursive call from within TransitionTo, just enqueue the
        // state transition so it can be handled below.
        if (queue.size() > 1) {
          LOG_DEBUG(TELEMETRY_CATEGORY_STATEMACHINE, fmt::format("Queue size [{}]: {}", name, queue.size()));
          return;
        }
      }

      // Repeatedly pop the front of the queue until empty.
      while (true) {
        StateTransition<T> st;
        {
          std::lock_guard<std::mutex> lock(queueMutex);
          st = queue.front();
        }

        const auto ns = stateStrings[st.newState];
        if (st.forceTransition || CanTransition(currentState, st.newState)) {
          LOG_DEBUG(TELEMETRY_CATEGORY_STATEMACHINE, fmt::format("Transition [{}]: {} -> {}", name, cs, ns));
          if (TransitionTo(st.newState)) {
            currentState = st.newState;
            LOG_VALUE(TELEMETRY_CATEGORY_STATEMACHINE, name, stateStrings[currentState]);
            lastChangeTimer.Reset();
          }
          else {
            LOG_WARNING(TELEMETRY_CATEGORY_STATEMACHINE, fmt::format("Failed state change [{}]: {} -> {}", name, cs, ns));
          }
        }
        else {
          LOG_ERROR(TELEMETRY_CATEGORY_STATEMACHINE, fmt::format("Invalid state change [{}]: {} -> {}", name, cs, ns));
        }

        {
          std::lock_guard<std::mutex> lock(queueMutex);
          queue.pop_front();
          if (queue.size() == 0) {
            return;
          }
        }
      }
    }

  private:
    T currentState;
    const T defaultState;
    frc::Timer lastChangeTimer;
    std::vector<Path<T>> allowed;
    const StateStrings stateStrings;

    // Queue for recursive state changes.
    std::mutex queueMutex;
    std::list<StateTransition<T>> queue;
  };

} // namespace Rebels
