/**
 * Copyright 2019 Kitchener-Waterloo Youth Robotics
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not
 * use this file except in compliance with the License. You may obtain a copy of
 * the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <algorithm>
#include <frc/Timer.h>

#include "Telemetry.h"
#include "StateMachines.h"

using namespace Rebels;

#define LOG_LONG_UPDATES
#define TELEMETRY_CATEGORY "State Machines"

bool compareFunc(Updateable* a, Updateable* b) {
  return (a->GetUpdatePriority() < b->GetUpdatePriority());
}

void Updateable::SetUpdatePriority(UpdateablePriority p) {
  updatePriority = p;
  StateMachines::UpdateOrder();
}

StateMachines::StateMachines() {
  backgroundThread = std::thread(&StateMachines::BackgroundThread, this);
}

void StateMachines::UpdateAll_() {
  if (updateOrder) {
    updateOrder = false;
    std::unique_lock write_lock(mutex);
    std::sort(stateMachines.begin(), stateMachines.end(), compareFunc);
  }

  std::shared_lock read_lock(mutex);

#ifdef LOG_LONG_UPDATES
  frc::Timer timerAll;
  timerAll.Start();
  timerAll.Reset();
#endif

  frc::Timer timerSingle;
  timerSingle.Start();
  for (const auto& entry : stateMachines) {
    timerSingle.Reset();
    entry->Update();
    entry->lastUpdateDuration = timerSingle.Get();
  }

#ifdef LOG_LONG_UPDATES
  if (timerAll.Get() >= 10_ms) {
    LOG_WARNING(TELEMETRY_CATEGORY, fmt::format("Update total duration {:0.2f}ms", timerAll.Get().convert<units::millisecond>().value()));
    for (const auto& entry : stateMachines) {
      LOG_WARNING(TELEMETRY_CATEGORY, fmt::format("Update '{}' duration {:0.2f}ms", entry->GetName(), entry->lastUpdateDuration.convert<units::millisecond>().value()));
    }
  }
#endif
}

void StateMachines::Register_(Updateable* sm) {
  std::unique_lock write_lock(mutex);
  stateMachines.push_back(sm);
  LOG_INFO(TELEMETRY_CATEGORY, fmt::format("Registered '{}'", sm->GetName()));
  updateOrder = true;
}

void StateMachines::Unregister_(Updateable* sm) {
  std::unique_lock write_lock(mutex);
  for (auto entry = stateMachines.begin(); entry != stateMachines.end(); entry++) {
    if (*entry == sm) {
      LOG_INFO(TELEMETRY_CATEGORY, fmt::format("Unregistered '{}'", (*entry)->GetName()));
      stateMachines.erase(entry);
      return;
    }
  }
}

void StateMachines::UpdateOrder_() {
  updateOrder = true;
}

void StateMachines::BackgroundThread() {
  backgroundShouldRun = true;
  frc::Wait(2_s); // wait to avoid races with the main thread

  LOG_INFO(TELEMETRY_CATEGORY, "Starting low priority background thread");

  if (backgroundLoopDelay < 20_ms) {
    backgroundLoopDelay = 20_ms;
  }

  while (backgroundShouldRun) {
    for (auto entry = stateMachines.begin(); entry != stateMachines.end() && backgroundShouldRun; entry++) {
      std::shared_lock read_lock(mutex);
      (*entry)->BackgroundUpdate();
    }
    pthread_yield();
    frc::Wait(backgroundLoopDelay);
  }

  LOG_INFO(TELEMETRY_CATEGORY, "Stopping low priority background thread");
}
