/**
 * Copyright 2019 Kitchener-Waterloo Youth Robotics
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not
 * use this file except in compliance with the License. You may obtain a copy of
 * the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <frc/DriverStation.h>
#include <frc/smartdashboard/SmartDashboard.h>
#include <frc/Timer.h>
#include <wpi/raw_ostream.h>
#include <fstream>
#include <hal/HALBase.h>

#include "Telemetry.h"

using namespace Rebels;

/**
 * Returns FPGA timestamp (in microseconds) with a fallback to zero when HAL
 * is not yet initialized and FPGA  is not available.
 *
 * This manages the situation when logging is invoked from within static methods,
 * e.g. Settings initialization.
 */
uint64_t GetSafeFPGATimestamp() {
  int32_t status = 0;
  const uint64_t time = HAL_GetFPGATime(&status);
  return status != 0 ? 0 : time;
}

static const constexpr std::string_view kWarnStr{ "[WARN]  " };
static const constexpr std::string_view kErrorStr{ "[ERROR] " };
static const constexpr std::string_view kDebugStr{ "[DEBUG] " };
static const constexpr std::string_view kInfoStr{ "[INFO]  " };

StaticTelemetry::StaticTelemetry() {
#ifdef TELEMETRY_FILESYSTEM
  drainQueueThread = std::thread(&StaticTelemetry::drainQueue, this);
#endif
}

void StaticTelemetry::Value(std::string_view category, std::string_view topic, double value) {
#ifdef TELEMETRY_SMARTDASHBOARD
  const auto name = fmt::format("{}/{}", category, topic);
  frc::SmartDashboard::PutNumber(name, value);
#endif

#ifdef TELEMETRY_FILESYSTEM
  Get()->metric(category, topic, value);
#endif
}

void StaticTelemetry::Value(std::string_view category, std::string_view topic, bool value) {
#ifdef TELEMETRY_SMARTDASHBOARD
  const auto name = fmt::format("{}/{}", category, topic);
  frc::SmartDashboard::PutBoolean(name, value);
#endif
}

void StaticTelemetry::Value(std::string_view category, std::string_view topic, std::string_view value) {
#ifdef TELEMETRY_SMARTDASHBOARD
  const auto name = fmt::format("{}/{}", category, topic);
  frc::SmartDashboard::PutString(name, value);
#endif
}

void StaticTelemetry::Info(std::string_view category, std::string_view message) {
  Get()->log(TelemetryLogLevel::Info, category, message);
}

void StaticTelemetry::Warning(std::string_view category, std::string_view message) {
  Get()->log(TelemetryLogLevel::Warning, category, message);
}

void StaticTelemetry::Error(std::string_view category, std::string_view message) {
  Get()->log(TelemetryLogLevel::Error, category, message);
}

void StaticTelemetry::Debug(std::string_view category, std::string_view message) {
  Get()->log(TelemetryLogLevel::Debug, category, message);
}

const std::string_view StaticTelemetry::formatLevel(TelemetryLogLevel level) const {
  return (
    level == TelemetryLogLevel::Warning
    ? kWarnStr
    : (level == TelemetryLogLevel::Debug
      ? kDebugStr
      : (level == TelemetryLogLevel::Error ? kErrorStr : kInfoStr)));
}

void StaticTelemetry::log(TelemetryLogLevel level, std::string_view category, std::string_view message) {
  const auto timestamp = GetSafeFPGATimestamp();

#ifdef TELEMETRY_CONSOLE
  wpi::outs() << fmt::format("{:>9L} ", timestamp) << formatLevel(level) << category << ": " << message << '\n';
#endif

#ifdef TELEMETRY_FILESYSTEM
  std::unique_ptr<TelemetryLogEntry> entry(new TelemetryLogEntry(timestamp, level, category, message));
  std::lock_guard<std::mutex> lock(queueLogsMutex);
  queueLogs.push_back(std::move(entry));
#endif
}

void StaticTelemetry::metric(std::string_view category, std::string_view topic, double value) {
#ifdef TELEMETRY_METRICS
  const auto timestamp = GetSafeFPGATimestamp();

#ifdef TELEMETRY_FILESYSTEM
  std::unique_ptr<TelemetryMetricEntry> entry(new TelemetryMetricEntry(timestamp, category, topic, value));
  std::lock_guard<std::mutex> lock(queueMetricsMutex);
  queueMetrics.push_back(std::move(entry));
#endif
#endif
}

#ifdef TELEMETRY_FILESYSTEM
void StaticTelemetry::drainQueue() {
  frc::Wait(5_s);
  const auto waitTime = 0.5_s;

  while (true) {
    const int matchNumber = frc::DriverStation::GetMatchNumber();

    const std::time_t t = std::time(0);
    const std::tm* now = std::localtime(&t);

    char filenameBase[256];
    const auto filenameBaseLength = sprintf(
      filenameBase, "/home/lvuser/telemetry_%d_%04d-%02d-%02d_%02d%02d%02d",
      matchNumber, now->tm_year + 1900, now->tm_mon + 1, now->tm_mday,
      now->tm_hour, now->tm_min, now->tm_sec);

    if (filenameBaseLength < 0) {
      wpi::outs() << "[ERROR] Telemetry: Failed to format telemetry filename\n";
      return;
    }

    const auto filenameLog = new char[filenameBaseLength + 5];
    std::strcpy(filenameLog, filenameBase);
    std::strcat(filenameLog, ".log");

    std::ofstream streamLog;
    streamLog.open(filenameLog, std::ios::out | std::ios::app);
    if (streamLog.fail()) {
      wpi::outs() << "[ERROR] Telemetry: Failed to open telemetry logs file '" << filenameLog << "'!\n";
      return;
    }

#ifdef TELEMETRY_METRICS
    const auto filenameCsv = new char[filenameBaseLength + 5];
    std::strcpy(filenameCsv, filenameBase);
    std::strcat(filenameCsv, ".csv");

    std::ofstream streamCsv;
    streamCsv.open(filenameCsv, std::ios::out | std::ios::app);
    if (streamCsv.fail()) {
      wpi::outs() << "[ERROR] Telemetry: Failed to open telemetry metrics file '" << filenameCsv << "'!\n";
      return;
    }
#endif

    while (matchNumber == frc::DriverStation::GetMatchNumber()) {
      pthread_yield();
      frc::Wait(waitTime);

      drainQueueLogs(streamLog);

#ifdef TELEMETRY_METRICS
      drainQueueMetrics(streamCsv);
#endif
    }

    streamLog.close();

#ifdef TELEMETRY_METRICS
    streamCsv.close();
#endif
  }
}

/**
 * Drains the log queue and pushes entries to the specified I/O stream.
 */
void StaticTelemetry::drainQueueLogs(std::ofstream& s) {
  std::vector<std::unique_ptr<TelemetryLogEntry>> local;
  {
    std::lock_guard<std::mutex> lock(queueLogsMutex);

    local.reserve(queueLogs.size());
    for (const auto& entry : queueLogs) {
      local.push_back(std::make_unique<TelemetryLogEntry>(*entry));
    }

    queueLogs.clear();
  }

  if (local.empty()) {
    return;
  }

  for (const auto& entry : local) {
    // {timestamp} [LEVEL] {category}: {message}
    s << fmt::format("{:>9L} ", entry->timestamp) << formatLevel(entry->level) << entry->category << ": " << entry->message << '\n';
  }

  s.flush();
}

/**
 * Drains the metrics queue and pushes entries to the specified I/O stream.
 */
void StaticTelemetry::drainQueueMetrics(std::ofstream& s) {
  std::vector<std::unique_ptr<TelemetryMetricEntry>> local;
  {
    std::lock_guard<std::mutex> lock(queueMetricsMutex);

    local.reserve(queueMetrics.size());
    for (const auto& entry : queueMetrics) {
      local.push_back(std::make_unique<TelemetryMetricEntry>(*entry));
    }

    queueMetrics.clear();
  }

  if (local.empty()) {
    return;
  }

  for (const auto& entry : local) {
    // {timestamp},{value},{category},{topic}
    s << entry->timestamp << ',' << entry->value << ',' << entry->category << ',' << entry->topic << '\n';
  }

  s.flush();
}
#endif
