/**
 *  ____ _____ ___ ____            ____      _          _
 * |___ \___  / _ \___ \          |  _ \ ___| |__   ___| |___
 *   __) | / / | | |__) |  _____  | |_) / _ \ '_ \ / _ \ / __|
 *  / __/ / /| |_| / __/  |_____| |  _ <  __/ |_) |  __/ \__ \
 * |_____/_/  \___/_____|         |_| \_\___|_.__/ \___|_|___/
 *
 *
 * Copyright 2019 Kitchener-Waterloo Youth Robotics
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not
 * use this file except in compliance with the License. You may obtain a copy of
 * the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#pragma once

#include <mutex>
#include <queue>
#include <string>
#include <thread>
#include <vector>
#include <fmt/core.h>

#include "Singleton.h"

#define TELEMETRY_SMARTDASHBOARD
#define TELEMETRY_FILESYSTEM
#define TELEMETRY_CONSOLE
 // #define TELEMETRY_METRICS

namespace Rebels {

  enum TelemetryLogLevel {
    None = 0, // no log output
    Info = 1,
    Warning = 2,
    Error = 3,
    Debug = 4,
  };

  class TelemetryLogEntry {
    friend class StaticTelemetry;
    TelemetryLogEntry(std::uint64_t timestamp, TelemetryLogLevel level, std::string_view category, std::string_view message) :
      timestamp(timestamp), level(level), category(category), message(message) {};

  private:
    const std::uint64_t timestamp;
    const TelemetryLogLevel level;
    const std::string category;
    const std::string message;
  };

  class TelemetryMetricEntry {
    friend class StaticTelemetry;
    TelemetryMetricEntry(std::uint64_t timestamp, std::string_view category, std::string_view topic, double value) :
      timestamp(timestamp), category(category), topic(topic), value(value) {};

  private:
    const std::uint64_t timestamp;
    const std::string category;
    const std::string topic;
    const double value;
  };

  class StaticTelemetry {
    DECLARE_SINGLETON(StaticTelemetry)

  public:
    StaticTelemetry();
    ~StaticTelemetry() = default;

    static void Value(std::string_view category, std::string_view topic, double value);
    static void Value(std::string_view category, std::string_view topic, bool value);
    static void Value(std::string_view category, std::string_view topic, std::string_view value);
    static void Info(std::string_view category, std::string_view message);
    static void Warning(std::string_view category, std::string_view message);
    static void Error(std::string_view category, std::string_view message);
    static void Debug(std::string_view category, std::string_view message);

  private:
    void log(TelemetryLogLevel level, std::string_view category, std::string_view message);
    void metric(std::string_view category, std::string_view topic, double value);
    const std::string_view formatLevel(TelemetryLogLevel level) const;

#ifdef TELEMETRY_FILESYSTEM
    void drainQueue();
    std::thread drainQueueThread;

    void drainQueueLogs(std::ofstream& s);
    void drainQueueMetrics(std::ofstream& s);

    std::mutex queueLogsMutex;
    std::vector<std::unique_ptr<TelemetryLogEntry>> queueLogs;

    std::mutex queueMetricsMutex;
    std::vector<std::unique_ptr<TelemetryMetricEntry>> queueMetrics;
#endif
  };

} // namespace Rebels

#define LOG_VALUE Rebels::StaticTelemetry::Value
#define LOG_INFO Rebels::StaticTelemetry::Info
#define LOG_WARNING Rebels::StaticTelemetry::Warning
#define LOG_ERROR Rebels::StaticTelemetry::Error
#define LOG_DEBUG Rebels::StaticTelemetry::Debug
