/**
 *  ____ _____ ___ ____            ____      _          _
 * |___ \___  / _ \___ \          |  _ \ ___| |__   ___| |___
 *   __) | / / | | |__) |  _____  | |_) / _ \ '_ \ / _ \ / __|
 *  / __/ / /| |_| / __/  |_____| |  _ <  __/ |_) |  __/ \__ \
 * |_____/_/  \___/_____|         |_| \_\___|_.__/ \___|_|___/
 *
 *
 * Copyright 2019 Kitchener-Waterloo Youth Robotics
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not
 * use this file except in compliance with the License. You may obtain a copy of
 * the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#pragma once

#include <units/time.h>

namespace Rebels {

  typedef int UpdateablePriority;

  static const constexpr UpdateablePriority CriticalUpdateablePriority = 10;
  static const constexpr UpdateablePriority HighUpdateablePriority = 20;
  static const constexpr UpdateablePriority MediumUpdateablePriority = 30;
  static const constexpr UpdateablePriority DefaultUpdateablePriority = 50;
  static const constexpr UpdateablePriority LowUpdateablePriority = 200;

  class StateMachines;

  class Updateable {
    friend class StateMachines;

  public:
    Updateable(std::string_view name, UpdateablePriority p = DefaultUpdateablePriority) :
      name(name), updatePriority(p) {}

    virtual ~Updateable() {}

    // Override to receive updates in the main thread. Time-sensitive.
    virtual void Update() = 0;

    // Override to receive updates in a background thread. Not time-sensitive.
    // Be sure not to block the main thread here.
    virtual void BackgroundUpdate() {}

    inline std::string_view GetName() const { return name; }
    inline UpdateablePriority GetUpdatePriority() const { return updatePriority; }
    void SetUpdatePriority(UpdateablePriority p);

  protected:
    const std::string_view name;

  private:
    UpdateablePriority updatePriority = DefaultUpdateablePriority;
    units::second_t lastUpdateDuration{ 0 };
  };

} // namespace Rebels
