/**
 * Copyright 2019 Kitchener-Waterloo Youth Robotics
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not
 * use this file except in compliance with the License. You may obtain a copy of
 * the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <asm/ioctls.h>
#include <asm/termbits.h>
#include <errno.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <unistd.h>
#include <wpi/fmt/raw_ostream.h>

extern "C" int tcflush(int fildes, int queue_selector);

#include "UsbSerial.h"
#include "Telemetry.h"

#define TELEMETRY_CATEGORY "Usb Serial"

UsbSerial::UsbSerial() {
  for (int i = 0; i < 10 && fd < 0; i++) {
    fd = TryOpen(i);
  }

  if (fd < 0) {
    LOG_ERROR(TELEMETRY_CATEGORY, "Failed to open any USB ports");
    return;
  }

  // set speed to 115,200 bps, 8n1 (no parity)
  SetSpeed(115200);
}

UsbSerial::~UsbSerial() {
  Close();
}

int UsbSerial::TryOpen(int port) {
  const auto portName = fmt::format("/dev/ttyUSB{:d}", port);
  const auto h = ::open(portName.c_str(), O_RDWR | O_NOCTTY | O_NDELAY);
  if (h < 0) {
    LOG_WARNING(TELEMETRY_CATEGORY, fmt::format("Failed to open USB port {}: {}", portName, strerror(errno)));
    return -1;
  }
  return h;
}

void UsbSerial::Close() {
  if (fd >= 0) {
    close(fd);
    fd = -1;
  }
}

int UsbSerial::SetSpeed(int baud) {
  // using Linux extension ...
  struct termios2 tio;

  ioctl(fd, TCGETS2, &tio);
  bzero(&tio, sizeof(struct termios2));

  tio.c_cflag = BOTHER;
  tio.c_cflag |= (CLOCAL | CREAD | CS8);  // 8 bit no hardware handshake

  tio.c_cflag &= ~CSTOPB;   // 1 stop bit
  tio.c_cflag &= ~CRTSCTS;  // No CTS
  tio.c_cflag &= ~PARENB;   // No Parity

#ifdef CNEW_RTSCTS
  tio.c_cflag &= ~CNEW_RTSCTS;  // no hw flow control
#endif

  tio.c_iflag &= ~(IXON | IXOFF | IXANY);  // no sw flow control

  tio.c_cc[VMIN] = 0;   // min chars to read
  tio.c_cc[VTIME] = 0;  // time in 1/10th sec wait

  tio.c_lflag &= ~(ICANON | ECHO | ECHOE | ISIG);
  // raw output mode
  tio.c_oflag &= ~OPOST;

  tio.c_ispeed = baud;
  tio.c_ospeed = baud;

  ioctl(fd, TCSETS2, &tio);

  Flush();

  if (fcntl(fd, F_SETFL, FNDELAY) < 0) {
    Close();
    return false;
  }

  ClearDTR();

  return 0;
}

void UsbSerial::Flush() {
  tcflush(fd, TCIFLUSH);
}

int UsbSerial::GetBytesReceived() {
  int bytes;
  if (ioctl(fd, FIONREAD, &bytes) < 0) {
    return 0;
  }
  return bytes;
}

int UsbSerial::Read(char* buffer, int count) {
  int cnt = read(fd, buffer, count);
  return cnt;
}

int UsbSerial::Write(const char* buffer, int count) {
  // write bytes
  return write(fd, buffer, count);
}

void UsbSerial::RaiseDTR() {
  if (fd >= 0) {
    uint32_t dtr_bit = TIOCM_DTR;
    if (ioctl(fd, TIOCMBIS, &dtr_bit) < 0) {
      LOG_ERROR(TELEMETRY_CATEGORY, fmt::format("Failed to raise DTR: {}", strerror(errno)));
    }
  }
}

void UsbSerial::ClearDTR() {
  if (fd >= 0) {
    uint32_t dtr_bit = TIOCM_DTR;
    if (ioctl(fd, TIOCMBIC, &dtr_bit) < 0) {
      LOG_ERROR(TELEMETRY_CATEGORY, fmt::format("Failed to clear DTR: {}", strerror(errno)));
    }
  }
}
