/**
 * Copyright 2019 Kitchener-Waterloo Youth Robotics
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not
 * use this file except in compliance with the License. You may obtain a copy of
 * the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <math.h>

#include "XboxTeleop.h"
#include "Settings.h"
#include "Telemetry.h"

using namespace Rebels;

#define TELEMETRY_CATEGORY "Teleop"

REGISTER_DOUBLE(DriveSpeedScale, "Drive/Speed scaling", 1.0)
REGISTER_DOUBLE(DriveTurnScaleMax, "Drive/Turn scaling max", 0.85)
REGISTER_DOUBLE(DriveTurnScaleMin, "Drive/Turn scaling min", 0.45)

void XboxTeleop::Periodic() {
  CheckStick();
  CheckShifters();
  UpdateRumble();
}

bool XboxTeleop::IsJoystickMoved() {
  const double moveTolerance = 0.5;
  return (
    std::fabs(driverController.GetLeftX()) > moveTolerance ||
    std::fabs(driverController.GetLeftY()) > moveTolerance ||
    std::fabs(driverController.GetRightX()) > moveTolerance ||
    std::fabs(driverController.GetRightY()) > moveTolerance);
}

void XboxTeleop::ClearButtonState() {
  LOG_INFO(TELEMETRY_CATEGORY, "Clearing button state");
  {
    const int cnt = driverController.GetButtonCount();
    for (int i = 1; i <= cnt; i++) {
      driverController.GetRawButtonPressed(i);
      driverController.GetRawButtonReleased(i);
    }
  }
#ifdef ENABLE_OPSTICK
  {
    const int cnt = opController.GetButtonCount();
    for (int i = 1; i <= cnt; i++) {
      opController.GetRawButtonPressed(i);
      opController.GetRawButtonReleased(i);
    }
  }
#endif
}

void XboxTeleop::CheckStick() {
  // only move the motors if some other system doesn't own it, like an auto
  if (!HasDrive()) {
    return;
  }

  // Defaults to left hand throttle and gears, right hand steering
#ifdef SWAP_CONTROLLER
  const double steer = driverController.GetLeftX();
  const double speed = driverController.GetRightY();
#else
  const double steer = -driverController.GetRightX();
  const double speed = -driverController.GetLeftY();
#endif

  // const double turnRange = GetDriveTurnScaleMax() - GetDriveTurnScaleMin();
  // const double turnScale = GetDriveTurnScaleMax() - (std::fabs(steer) * turnRange);
  // (*drive)->ArcadeDrive(speed * GetDriveSpeedScale(), steer * turnScale, false);
  (*drive)->ArcadeDrive(speed, steer * GetDriveTurnScaleMax(), true);
}

void XboxTeleop::CheckShifters() {
#ifdef ENABLE_SHIFTERS
  if (!HasDrive()) {
    return;
  }

#ifdef SWAP_CONTROLLER
  if (driverController.GetRightStickButtonPressed()) {
#else
  if (driverController.GetXButtonPressed()) {
#endif
    if ((*drive)->IsHighGear()) {
      LOG_INFO(TELEMETRY_CATEGORY, "Driver shifting low");
      (*drive)->ShiftLow();
    }
    else {
      LOG_INFO(TELEMETRY_CATEGORY, "Driver shifting high");
      (*drive)->ShiftHigh();
    }
  }
#endif
}

void XboxTeleop::Rumble() {
  driverController.SetRumble(frc::XboxController::kLeftRumble, 1);
  driverController.SetRumble(frc::XboxController::kRightRumble, 1);
  rumbleCount = 15;
}

void XboxTeleop::UpdateRumble() {
  if (rumbleCount > 0) {
    if (--rumbleCount == 0) {
      driverController.SetRumble(frc::XboxController::kLeftRumble, 0);
      driverController.SetRumble(frc::XboxController::kRightRumble, 0);
    }
  }
}
