#include "Climber.h"
#include "Settings.h"
#include "Telemetry.h"
#include "Limelight.h"

using namespace Rebels;

#define TELEMETRY_CATEGORY "Climber"

// https://docs.ctre-phoenix.com/en/latest/ch14_MCSensor.html#confirm-sensor-resolution-velocity
// Quadrature Encoder
const double TALONSRX_QUADENC_UNITS_PER_REVOLUTION = 4096.0;
const double THRESHOLD_AT_POSITION = 0.1;

REGISTER_BOOL(ClimberVerboseDashboard, "Climber/Verbose", false)
REGISTER_DOUBLE(ClimberWinchF, "Climber/Winch PID/F", 0)
REGISTER_DOUBLE(ClimberWinchP, "Climber/Winch PID/P", 1.0)
REGISTER_DOUBLE(ClimberWinchI, "Climber/Winch PID/I", 0)
REGISTER_DOUBLE(ClimberWinchD, "Climber/Winch PID/D", 0)
//REGISTER_DOUBLE(ClimberWinchLoopError, "Climber: Winch PID - Error Abs", 1028)
//REGISTER_DOUBLE(ClimberWinchRamp, "Climber: Winch Ramp Time", 0.2)
REGISTER_DOUBLE(ClimberWinchPositionExtended, "Climber/Position Extended", 3.7)
REGISTER_DOUBLE(ClimberWinchPositionRetracted, "Climber/Position Retracted", -0.15)
REGISTER_DOUBLE(ClimberWinchPositionIntermediate, "Climber/Position Intermediate", 1.5)
REGISTER_DOUBLE(ClimberWinchPositionSecondBarExtended, "Climber/SecondBar Extended", 3.9)
REGISTER_DOUBLE(ClimberWinchPositionLastBar, "Climber/Position Last Bar", 2.0);

REGISTER_DOUBLE(ClimberWinchAutoHomeSpeed, "Climber/AutoHome/Speed", -0.4)
REGISTER_DOUBLE(ClimberWinchAutoHomeCurrentThreshold, "Climber/AutoHome/Current Threshold", 20.0)
REGISTER_DOUBLE(ClimberWinchAutoHomeTimeout, "Climber/AutoHome/Timeout", 5.0)
REGISTER_DOUBLE(ClimberWinchPositionReadout, "Climber/PositionReadout", 0.0)

REGISTER_DOUBLE(ClimberWinchCurrentOutput, "Climber/CurrentOutput", 0)



Climber::Climber() : StateMachine("Climber", ClimberStateStrings) {
  AddTransition(CLIMBER_INIT, CLIMBER_IDLE);

  // First bar and reset transitions.
  AddTransition(CLIMBER_IDLE, CLIMBER_EXTENDING);
  AddTransition(CLIMBER_EXTENDING, CLIMBER_EXTENDED);
  AddTransition(CLIMBER_EXTENDED, CLIMBER_RETRACTING);
  AddTransition(CLIMBER_RETRACTING, CLIMBER_RETRACTED);
  AddTransition(CLIMBER_RETRACTED, CLIMBER_EXTENDING);

  // Homing transitions.
  AddTransition(CLIMBER_IDLE, CLIMBER_HOMING);
  AddTransition(CLIMBER_HOMING, CLIMBER_HOMED);
  AddTransition(CLIMBER_HOMED, CLIMBER_IDLE);
  AddTransition(CLIMBER_ERROR, CLIMBER_IDLE);
  AddTransition(CLIMBER_EXTENDED, CLIMBER_HOMING);
  AddTransition(CLIMBER_RETRACTED, CLIMBER_HOMING);
  AddTransition(CLIMBER_HOMING, CLIMBER_IDLE);

  // Second bar transitions.
  AddTransition(CLIMBER_RETRACTED, CLIMBER_EXTENDING_NEXT_BAR);
  AddTransition(CLIMBER_EXTENDING_NEXT_BAR, CLIMBER_EXTENDED_NEXT_BAR);
  AddTransition(CLIMBER_EXTENDED_NEXT_BAR, CLIMBER_RETRACTING_NEXT_BAR);
  AddTransition(CLIMBER_RETRACTING_NEXT_BAR, CLIMBER_RETRACTED_NEXT_BAR);
  AddTransition(CLIMBER_RETRACTED_NEXT_BAR, CLIMBER_EXTENDING_NEXT_BAR);

  // Third bar transitions.
  AddTransition(CLIMBER_EXTENDED_NEXT_BAR, CLIMBER_RETRACTING_LAST_BAR);
  AddTransition(CLIMBER_RETRACTING_LAST_BAR, CLIMBER_RETRACTED_LAST_BAR);

  //#ifdef ENABLE_SHOOTER
  shooterRef = Shooter::Get();
  //#endif
}

bool Climber::TransitionTo(ClimberState newState) {
  switch (newState) {
  case CLIMBER_INIT:
  case CLIMBER_IDLE:
    StopMotors();
    return true;
  case CLIMBER_EXTENDING:
    ZeroWinchPosition();
    SetWinchMotorPosition(GetClimberWinchPositionExtended());
    FireSolenoid(frc::DoubleSolenoid::kForward);
    return true;
  case CLIMBER_EXTENDED:
    StopMotors();
    FireSolenoid(frc::DoubleSolenoid::kOff);
    return true;
  case CLIMBER_RETRACTING:
    SetWinchMotorPosition(GetClimberWinchPositionRetracted());
    FireSolenoid(frc::DoubleSolenoid::kOff);
    return true;
  case CLIMBER_RETRACTED:
    FireSolenoid(frc::DoubleSolenoid::kReverse);
    return true;
  case CLIMBER_EXTENDING_NEXT_BAR:
    extendingWait = frc::Timer::GetFPGATimestamp();
    //Un hook from the one bar and back around up to the other one.
    FireSolenoid(frc::DoubleSolenoid::kForward);
    SetWinchMotorPosition(GetClimberWinchPositionIntermediate());
    return true;
  case CLIMBER_EXTENDED_NEXT_BAR:
    isClearOfBar = false;
    FireSolenoid(frc::DoubleSolenoid::kForward);
    return true;
  case CLIMBER_RETRACTING_NEXT_BAR:
    SetWinchMotorPosition(GetClimberWinchPositionRetracted());
    FireSolenoid(frc::DoubleSolenoid::kOff);
    return true;
  case CLIMBER_RETRACTED_NEXT_BAR:
    isOnSecondBar = true;
    FireSolenoid(frc::DoubleSolenoid::kReverse);
    return true;
  case CLIMBER_RETRACTING_LAST_BAR:
    SetWinchMotorPosition(GetClimberWinchPositionLastBar());
    FireSolenoid(frc::DoubleSolenoid::kOff);
    return true;
  case CLIMBER_RETRACTED_LAST_BAR:
    SetWinchMotorPower(0.0);
    return true;
  case CLIMBER_HOMING:
    FireSolenoid(frc::DoubleSolenoid::kReverse);
    homingStartTime = frc::Timer::GetFPGATimestamp();
    SetWinchMotorPower(GetClimberWinchAutoHomeSpeed());
    return true;
  case CLIMBER_HOMED:
    // Report position difference?
    ZeroWinchPosition();
    MakeHomedSound();
    homedTime = frc::Timer::GetFPGATimestamp();
    return true;
  case CLIMBER_ERROR:
    StopMotors();
    // Flash LED's
    // Make sound.
    // Show Dialog or something.
    return true;
  }
  return false;
}

void Climber::Update() {

  SetClimberWinchPositionReadout(GetWinchMotorPosition());
  SetClimberWinchCurrentOutput(GetWinchCurrentDraw());

  switch (GetCurrentState()) {
  case CLIMBER_INIT:
    Init();
    ForceState(CLIMBER_IDLE);
    break;

  case CLIMBER_IDLE:
    break;

  case CLIMBER_EXTENDING:
    if (fabs(GetWinchMotorPosition() - GetClimberWinchPositionExtended()) <= THRESHOLD_AT_POSITION) {
      ChangeState(CLIMBER_EXTENDED);
    }
    break;

  case CLIMBER_EXTENDED:
    break;

  case CLIMBER_RETRACTING:
    if (fabs(GetWinchMotorPosition() - GetClimberWinchPositionRetracted()) <= THRESHOLD_AT_POSITION) {
      ChangeState(CLIMBER_RETRACTED);
    }
    break;

  case CLIMBER_RETRACTED:
    break;

  case CLIMBER_EXTENDING_NEXT_BAR:
    if (!isClearOfBar)
    {
      if (fabs(GetWinchMotorPosition() - GetClimberWinchPositionIntermediate()) <= THRESHOLD_AT_POSITION) {
        if (extendingWait + units::second_t(1.0) <= frc::Timer::GetFPGATimestamp()) {
          FireSolenoid(frc::DoubleSolenoid::kReverse);
          SetWinchMotorPosition(GetClimberWinchPositionSecondBarExtended());
          isClearOfBar = true;
        }
      }
    }
    else
    {
      if ((fabs(GetWinchMotorPosition() - GetClimberWinchPositionSecondBarExtended()) <= THRESHOLD_AT_POSITION) &&
        (extendingWait + units::second_t(3.0) <= frc::Timer::GetFPGATimestamp())) {
        ChangeState(CLIMBER_EXTENDED_NEXT_BAR);
      }
    }
    break;

  case CLIMBER_EXTENDED_NEXT_BAR:
    break;

  case CLIMBER_RETRACTING_NEXT_BAR:
    if (fabs(GetWinchMotorPosition() - GetClimberWinchPositionRetracted()) <= THRESHOLD_AT_POSITION) {  //THis thresh value was -0.4
      ChangeState(CLIMBER_RETRACTED_NEXT_BAR);
    }
    break;

  case CLIMBER_RETRACTED_NEXT_BAR:
    break;

  case CLIMBER_RETRACTING_LAST_BAR:
    if (fabs(GetWinchMotorPosition() - GetClimberWinchPositionLastBar()) <= THRESHOLD_AT_POSITION) {
      ChangeState(CLIMBER_RETRACTED_LAST_BAR);
    }
    break;

  case CLIMBER_RETRACTED_LAST_BAR:
    break;

  case CLIMBER_HOMING:
    // Might need to set the motor power here on each loop iteration.
    if (GetWinchCurrentDraw() > GetClimberWinchAutoHomeCurrentThreshold()) {
      StopMotors();
      ChangeState(CLIMBER_HOMED);
    }
    if (homingStartTime + units::second_t(GetClimberWinchAutoHomeTimeout()) <= frc::Timer::GetFPGATimestamp()) {
      StopMotors();
      ForceState(CLIMBER_ERROR);
    }
    break;

  case CLIMBER_HOMED:
    if (Limelight::GetLedState() == LimelightLedState::Off)
      Limelight::SetLedState(LimelightLedState::Flashing);

    if (homedTime + units::second_t(0.5) <= frc::Timer::GetFPGATimestamp()) {
      if (Limelight::GetLedState() == LimelightLedState::Flashing)
        Limelight::SetLedState(LimelightLedState::Off);
      ChangeState(CLIMBER_IDLE);
    }
    break;

  case CLIMBER_ERROR:
    StopMotors();
    break;
  }
}

void Climber::BackgroundUpdate() {
  if (!GetClimberVerboseDashboard()) {
    return;
  }


  LOG_VALUE(TELEMETRY_CATEGORY, "Winch Motor Position", GetWinchMotorPosition());
  LOG_VALUE(TELEMETRY_CATEGORY, "Winch Motor Current Output", GetWinchCurrentDraw());
}

void Climber::AutoHome() {
  const auto state = GetCurrentState();
  if (state == CLIMBER_IDLE) {
    LOG_INFO(TELEMETRY_CATEGORY, "AutoHoming");
    ChangeState(CLIMBER_HOMING);
  }
}

void Climber::Extend() {
  const auto state = GetCurrentState();
  if (state == CLIMBER_IDLE || state == CLIMBER_RETRACTED) {
    LOG_INFO(TELEMETRY_CATEGORY, "Extending");
    ChangeState(CLIMBER_EXTENDING);
  }
}

void Climber::Retract() {
  const auto state = GetCurrentState();
  if (state == CLIMBER_EXTENDED) {
    LOG_INFO(TELEMETRY_CATEGORY, "Retracting");
    ChangeState(CLIMBER_RETRACTING);
  }
}

void Climber::NextBarExtend() {
  const auto state = GetCurrentState();
  if (state == CLIMBER_RETRACTED || state == CLIMBER_RETRACTED_NEXT_BAR) {
    LOG_INFO(TELEMETRY_CATEGORY, "Extending(next bar)");
    ChangeState(CLIMBER_EXTENDING_NEXT_BAR);
  }
}

void Climber::NextBarRetract() {
  const auto state = GetCurrentState();
  if (isOnSecondBar && state == CLIMBER_EXTENDED_NEXT_BAR) {
    LOG_INFO(TELEMETRY_CATEGORY, "Retracting(last bar)");
    ChangeState(CLIMBER_RETRACTING_LAST_BAR);
  }
  if (state == CLIMBER_EXTENDED_NEXT_BAR) {
    LOG_INFO(TELEMETRY_CATEGORY, "Retracting(next bar)");
    ChangeState(CLIMBER_RETRACTING_NEXT_BAR);
  }
}

// void Climber::ReversePneumatics() {
// #ifdef ENABLE_CLIMBER
//   FireSolenoid(frc::DoubleSolenoid::kReverse);
// #endif
// }

void Climber::Init() {
  LOG_INFO(TELEMETRY_CATEGORY, "Initializing");

  isOnSecondBar = false;

#ifdef ENABLE_CLIMBER
  const auto timeout = TIMEOUT.convert<units::ms>().value();
  winchMotor.ConfigFactoryDefault();
  winchMotor.SetNeutralMode(NeutralMode::Brake);

  // select local feedback source and zero sensor
  winchMotor.ConfigSelectedFeedbackSensor(ctre::phoenix::motorcontrol::FeedbackDevice::QuadEncoder, 0, timeout);
  winchMotor.SetSensorPhase(true);
  winchMotor.GetSensorCollection().SetQuadraturePosition(0, timeout);
  //winchMotor.SetSelectedSensorPosition(winchMotor.GetSelectedSensorPosition(), 0, timeout);
  //winchMotor.SetSelectedSensorPosition(winchMotor.GetSensorCollection().GetPulseWidthPosition(), 0, timeout);

  // Set up peak and nominal output.
  winchMotor.ConfigNominalOutputForward(0, timeout);
  winchMotor.ConfigNominalOutputReverse(0, timeout);
  winchMotor.ConfigPeakOutputForward(1, timeout);
  winchMotor.ConfigPeakOutputReverse(-1, timeout);

  // Set up PID.
  winchMotor.Config_kF(0, GetClimberWinchF(), timeout);
  winchMotor.Config_kP(0, GetClimberWinchP(), timeout);
  winchMotor.Config_kI(0, GetClimberWinchI(), timeout);
  winchMotor.Config_kD(0, GetClimberWinchD(), timeout);
  //winchMotor.ConfigClosedloopRamp(GetClimberWinchRamp());
  //winchMotor.ConfigAllowableClosedloopError(0, GetClimberWinchLoopError());
#endif

  StopMotors();
}

void Climber::StopMotors() {
#ifdef ENABLE_CLIMBER
  winchMotor.Set(ControlMode::PercentOutput, 0);
#endif
}

void Climber::SetWinchMotorPower(double power) {
#ifdef ENABLE_CLIMBER
  winchMotor.Set(ControlMode::PercentOutput, power);
#endif
}

void Climber::SetWinchMotorPosition(double position) {
#ifdef ENABLE_CLIMBER
  winchMotor.Set(ControlMode::Position, position * TALONSRX_QUADENC_UNITS_PER_REVOLUTION);
#endif
}

double Climber::GetWinchMotorPosition() {
#ifdef ENABLE_CLIMBER
  return winchMotor.GetSelectedSensorPosition(0) / TALONSRX_QUADENC_UNITS_PER_REVOLUTION;
#endif
}

double Climber::GetWinchCurrentDraw() {
#ifdef ENABLE_CLIMBER
  return fabs(winchMotor.GetStatorCurrent());
#endif
}

void Climber::ZeroWinchPosition() {
#ifdef ENABLE_CLIMBER
  winchMotor.GetSensorCollection().SetQuadraturePosition(0, TIMEOUT.convert<units::ms>().value());
#endif
}

void Climber::FireSolenoid(frc::DoubleSolenoid::Value value) {
#ifdef ENABLE_CLIMBER
#ifdef ENABLE_COMPRESSOR
  solenoid.Set(value);
#endif
#endif
}

void Climber::MakeHomedSound() {
#ifdef ENABLE_SHOOTER
  shooterRef->MakeSound();
#endif
}
