#pragma once
#include <ctre/Phoenix.h>
#include <frc/DoubleSolenoid.h>

#include "Constants.h"
#include "Settings.h"
#include "Singleton.h"
#include "StateMachine.h"
#include "Shooter.h"

namespace Rebels {

  enum ClimberState {
    CLIMBER_INIT,
    CLIMBER_IDLE,
    CLIMBER_EXTENDING,
    CLIMBER_EXTENDED,
    CLIMBER_RETRACTING,
    CLIMBER_RETRACTED,
    CLIMBER_EXTENDING_NEXT_BAR,
    CLIMBER_EXTENDED_NEXT_BAR,
    CLIMBER_RETRACTING_NEXT_BAR,
    CLIMBER_RETRACTED_NEXT_BAR,
    CLIMBER_RETRACTING_LAST_BAR,
    CLIMBER_RETRACTED_LAST_BAR,
    CLIMBER_HOMING,
    CLIMBER_HOMED,
    CLIMBER_ERROR
  };

  const StateMachine<ClimberState>::StateStrings ClimberStateStrings = {
    "Init",
    "Idle",
    "Extending",
    "Extended",
    "Retracting",
    "Retracted",
    "Extending Next Bar",
    "Extended Next Bar",
    "Retracting Next Bar",
    "Retracted Next Bar",
    "Retracting Last Bar",
    "Retracted Last Bar",
    "Homing",
    "Homed",
    "Error"
  };

  class Climber : public StateMachine<ClimberState> {
    DECLARE_SINGLETON(Climber)

  public:
    virtual ~Climber() = default;
    void Reset() { ForceState(CLIMBER_INIT); }

  public:
    virtual void Update() override;
    virtual void BackgroundUpdate() override;

    // Actions
    void AutoHome();
    void Extend();
    void Retract();
    void NextBarExtend();
    void NextBarRetract();
    //void ReversePneumatics();

  protected:
    virtual bool TransitionTo(ClimberState newState) override;

  private:
    Climber();
    void Init();
    void StopMotors();
    void SetWinchMotorPower(double power);
    void SetWinchMotorPosition(double position);
    double GetWinchMotorPosition();
    double GetWinchCurrentDraw();
    void ZeroWinchPosition();
    void FireSolenoid(frc::DoubleSolenoid::Value value);
    void MakeHomedSound();

    bool isClearOfBar = false;
    bool isOnSecondBar = false;
    units::second_t homingStartTime{ 0_s };
    units::second_t homedTime{ 0_s };
    units::second_t extendingWait{ 0_s };

  private:
#ifdef ENABLE_CLIMBER
    TalonSRX winchMotor{ CLIMBER_MOTOR_ID };
#ifdef ENABLE_COMPRESSOR
    frc::DoubleSolenoid solenoid{ COMPRESSOR_MODULE_TYPE, CLIMBER_FORWARD_PCM_ID, CLIMBER_BACKWARD_PCM_ID };
#endif
#endif
    Shooter* shooterRef = nullptr;
  };
}  // namespace Rebels




/*
Climber State: Idle
Winch Position: 3.4
Current Draw: 15
[HomeButton]
___________________
          [EditableToggle?]
Have this all in a table?
REGISTER_BOOL(ClimberVerboseDashboard, "Climber: Verbose", false)????
{ {Group them if possible.}
  ClimberWinchF : 0
  ClimberWinchP : 1.0
  ClimberWinchI : 0
  ClimberWinchD : 0
}
{
  ClimberWinchPositionExtended : 3.4
  ClimberWinchPositionRetracted : 0.2}
{
  ClimberWinchAutoHomeSpeed : -0.4
  ClimberWinchAutoHomeCurrentThreshold 20.0
  ClimberWinchAutoHomeTimeout : 5.0
}
*/