#include "Feeder.h"
#include "Settings.h"
#include "Telemetry.h"

using namespace Rebels;

#define TELEMETRY_CATEGORY "Feeder"

REGISTER_BOOL(FeederVerboseDashboard, "Feeder/Verbose", false)
REGISTER_BOOL(FeederBeamBreak, "Feeder/Beam", false)
REGISTER_DOUBLE(FeederStage1LeftRamp, "Feeder/Stage 1/Left Ramp Time", 0.4)
REGISTER_DOUBLE(FeederStage1LeftPowerIn, "Feeder/Stage 1/Left Power In", 1.0)
REGISTER_DOUBLE(FeederStage1LeftPowerOut, "Feeder/Stage 1/Left Power Out", -1.0)
REGISTER_DOUBLE(FeederStage1LeftPowerPre, "Feeder/Stage 1/Left Power PreShoot", -0.5)
REGISTER_DOUBLE(FeederStage1RightRamp, "Feeder/Stage 1/Right Ramp Time", 0.4)
REGISTER_DOUBLE(FeederStage1RightPowerIn, "Feeder/Stage 1/Right Power In", -1.0)
REGISTER_DOUBLE(FeederStage1RightPowerOut, "Feeder/Stage 1/Right Power Out", 1.0)
REGISTER_DOUBLE(FeederStage1RightPowerPre, "Feeder/Stage 1/Right Power PreShoot", -0.5)
REGISTER_DOUBLE(FeederStage2Ramp, "Feeder/Stage 1/Ramp Time", 0.4)
REGISTER_DOUBLE(FeederStage2PowerIn, "Feeder/Stage 2/Power In", 1.0)
REGISTER_DOUBLE(FeederStage2PowerOut, "Feeder/Stage 2/Power Out", -1.0)
REGISTER_DOUBLE(FeederStage2PowerPre, "Feeder/Stage 2/Power PreShoot", -0.3)
REGISTER_DOUBLE(FeederStage1Delay, "Feeder/Stage 1/Start Delay Time", 1)
REGISTER_DOUBLE(FeederStage1LeftPowerFeed, "Feeder/Stage 1/Left Power Feed", -0.4)
REGISTER_DOUBLE(FeederStage1RightPowerFeed, "Feeder/Stage 1/Right Power Feed", 0.4)

Feeder::Feeder() : StateMachine("Feeder", FeederStrings) {
  AddTransition(FEEDER_INIT, FEEDER_IDLE);
  AddTransition(FEEDER_IDLE, FEEDER_IDLE);
  AddTransition(FEEDER_IDLE, FEEDER_INTAKING);
  AddTransition(FEEDER_IDLE, FEEDER_EJECTING);
  AddTransition(FEEDER_IDLE, FEEDER_FEEDING);
  AddTransition(FEEDER_INTAKING, FEEDER_INTAKING);
  AddTransition(FEEDER_INTAKING, FEEDER_IDLE);
  AddTransition(FEEDER_INTAKING, FEEDER_FEEDING);
  AddTransition(FEEDER_INTAKING, FEEDER_EJECTING);
  AddTransition(FEEDER_EJECTING, FEEDER_EJECTING);
  AddTransition(FEEDER_EJECTING, FEEDER_IDLE);
  AddTransition(FEEDER_EJECTING, FEEDER_INTAKING);
  AddTransition(FEEDER_EJECTING, FEEDER_FEEDING);
  AddTransition(FEEDER_FEEDING, FEEDER_FEEDING);
  AddTransition(FEEDER_FEEDING, FEEDER_IDLE);
  AddTransition(FEEDER_FEEDING, FEEDER_INTAKING);
  AddTransition(FEEDER_FEEDING, FEEDER_EJECTING);
}

void Feeder::IsJogging(bool jogging)
{
  isJogging = jogging;
}

bool Feeder::TransitionTo(FeederState newState) {
  switch (newState) {
  case FEEDER_INIT:
    StopMotors();
    return true;
  case FEEDER_IDLE:
    if (beamInput.Get()) {
      SetStage1MotorPower(GetFeederStage1LeftPowerPre(), GetFeederStage1RightPowerPre());
      SetStage2MotorPower(GetFeederStage2PowerPre());
      idleRev = true;
      stage1Delay = frc::Timer::GetFPGATimestamp() + units::second_t(GetFeederStage1Delay());
    }
    else
      StopMotors();
    return true;
  case FEEDER_INTAKING:
    isJogging = false;
    SetStage1MotorPower(GetFeederStage1LeftPowerIn(), GetFeederStage1RightPowerIn());
    SetStage2MotorPower(0);
    return true;
  case FEEDER_EJECTING:
    isJogging = false;
    SetStage1MotorPower(GetFeederStage1LeftPowerOut(), GetFeederStage1RightPowerOut());
    SetStage2MotorPower(GetFeederStage2PowerOut());
    return true;
  case FEEDER_FEEDING:
    // slow down/stop stage 1 motors to improve shooting of 2 balls,
    // reducing a chance of in air collision
    isJogging = false;
    SetStage1MotorPower(GetFeederStage1LeftPowerPre(), GetFeederStage1RightPowerPre());
    SetStage2MotorPower(GetFeederStage2PowerIn());
    stage1Rev = true;
    stage1Delay = frc::Timer::GetFPGATimestamp() + units::second_t(GetFeederStage1Delay());
    return true;
  }
  return false;
}

void Feeder::Update() {
#ifdef ENABLE_FEEDER
  stage1MotorLeft.Set(ControlMode::PercentOutput, stage1MotorLeftPower);
  stage1MotorRight.Set(ControlMode::PercentOutput, stage1MotorRightPower);
  stage2Motor.Set(ControlMode::PercentOutput, stage2MotorPower);
#endif
  SetFeederBeamBreak(beamInput.Get());
  switch (GetCurrentState()) {
  case FEEDER_INIT:
    Init();
    ForceState(FEEDER_IDLE);
    break;

  case FEEDER_IDLE:
    if (idleRev) {
      if (!beamInput.Get() || frc::Timer::GetFPGATimestamp() >= stage1Delay) {
        StopMotors();
        idleRev = false;
      }
    }
    break;

  case FEEDER_INTAKING:
  case FEEDER_EJECTING:
    break;

  case FEEDER_FEEDING:
    if (stage1Rev) {
      if (!beamInput.Get()) {
        stage1Rev = false;
        SetStage1MotorPower(GetFeederStage1LeftPowerFeed(), GetFeederStage1RightPowerFeed());
      }
      else if (frc::Timer::GetFPGATimestamp() >= stage1Delay) {
        stage1Rev = false;
        SetStage1MotorPower(0, 0);
      }
    }
    break;
  }
}

void Feeder::BackgroundUpdate() {
  if (!GetFeederVerboseDashboard()) {
    return;
  }

  LOG_VALUE(TELEMETRY_CATEGORY, "Stage 1/Left Motor Power", stage1MotorLeftPower);
  LOG_VALUE(TELEMETRY_CATEGORY, "Stage 1/Right Motor Power", stage1MotorRightPower);
  LOG_VALUE(TELEMETRY_CATEGORY, "Stage 2/Motor Power", stage2MotorPower);
}

void Feeder::Intake() {
  if (!IsIntaking()) {
    LOG_INFO(TELEMETRY_CATEGORY, "Intaking");
    ChangeState(FEEDER_INTAKING);
  }
}

void Feeder::Eject() {
  if (!IsEjecting()) {
    LOG_INFO(TELEMETRY_CATEGORY, "Ejecting");
    ChangeState(FEEDER_EJECTING);
  }
}

void Feeder::Feed() {
  if (!IsFeeding()) {
    LOG_INFO(TELEMETRY_CATEGORY, "Feeding");
    ChangeState(FEEDER_FEEDING);
  }
}

void Feeder::Stop() {
  if (GetCurrentState() != FEEDER_IDLE) {
    LOG_INFO(TELEMETRY_CATEGORY, "Stopping");
    ChangeState(FEEDER_IDLE);
  }
}

bool Feeder::IsIntaking() const {
  return (GetCurrentState() == FEEDER_INTAKING);
}

bool Feeder::IsEjecting() const {
  return (GetCurrentState() == FEEDER_EJECTING);
}

bool Feeder::IsFeeding() const {
  return (GetCurrentState() == FEEDER_FEEDING);
}

void Feeder::Init() {
  LOG_INFO(TELEMETRY_CATEGORY, "Initializing");

#ifdef ENABLE_FEEDER
  InitMotor(stage1MotorLeft, GetFeederStage1LeftRamp());
  InitMotor(stage1MotorRight, GetFeederStage1RightRamp());
  InitMotor(stage2Motor, GetFeederStage2Ramp());
#endif

  StopMotors();
}

void Feeder::InitMotor(VictorSPX& motor, const double ramp) {
  motor.ConfigFactoryDefault();
  motor.ConfigClosedloopRamp(ramp);
  motor.ConfigOpenloopRamp(ramp);
  motor.SetNeutralMode(motorcontrol::NeutralMode::Brake);
}

void Feeder::StopMotors() {
  SetStage1MotorPower(0, 0);
  SetStage2MotorPower(0);
}

void LogMotorPowerAdjustment(const double value, const double current, const std::string_view name) {
  if (value > 0.001 && current < 0.001) {
    LOG_INFO(TELEMETRY_CATEGORY, fmt::format("Initiating {} forward motion", name));
  }

  if (value < -0.001 && current > -0.001) {
    LOG_INFO(TELEMETRY_CATEGORY, fmt::format("Initiating {} backward motion", name));
  }
}

void Feeder::SetStage1MotorPower(const double powerLeft, const double powerRight) {
  LogMotorPowerAdjustment(powerLeft, stage1MotorLeftPower, "stage 1 left");
  LogMotorPowerAdjustment(powerRight, stage1MotorRightPower, "stage 1 right");

  stage1MotorLeftPower = powerLeft;
  stage1MotorRightPower = powerRight;
}

void Feeder::SetStage2MotorPower(double power) {
  LogMotorPowerAdjustment(power, stage2MotorPower, "stage 2");

  stage2MotorPower = power;
}
