#pragma once
#include <ctre/Phoenix.h>
#include <frc/DigitalInput.h>

#include "Constants.h"
#include "Settings.h"
#include "Singleton.h"
#include "StateMachine.h"

namespace Rebels {

  enum FeederState {
    FEEDER_INIT,
    FEEDER_IDLE,
    FEEDER_INTAKING,
    FEEDER_EJECTING,
    FEEDER_FEEDING,
  };

  const StateMachine<FeederState>::StateStrings FeederStrings = {
    "Init",
    "Idle",
    "Intaking", // stage-1 motors are rolling forwards, stage-2 is idle
    "Ejecting", // stage-1 and stage-2 motors are rolling backwards
    "Feeding",  // stage-1 and stage-2 motors are rolling forwards
  };

  class Feeder : public StateMachine<FeederState> {
    DECLARE_SINGLETON(Feeder)

  public:
    virtual ~Feeder() = default;
    void Reset() { ForceState(FEEDER_INIT); }

  public:
    virtual void Update() override;
    virtual void BackgroundUpdate() override;

    // Actions
    void Intake();
    void Eject();
    void Feed();
    void Stop();

    bool IsIntaking() const;
    bool IsFeeding() const;
    bool IsEjecting() const;
    virtual void IsJogging(bool);

  protected:
    virtual bool TransitionTo(FeederState newState) override;

  private:
    Feeder();
    void Init();
    void InitMotor(VictorSPX& motor, const double ramp);
    void StopMotors();
    void SetStage1MotorPower(const double powerLeft, const double powerRight);
    void SetStage2MotorPower(const double power);

  private:
    double stage1MotorLeftPower{ 0 };
    double stage1MotorRightPower{ 0 };
    double stage2MotorPower{ 0 };
    units::second_t stage1Delay{ 0.0_s };
    bool stage1Rev{ false };
	bool idleRev{ false };
    bool isJogging{ false };
    frc::DigitalInput beamInput{ 9 };

#ifdef ENABLE_FEEDER
    VictorSPX stage1MotorLeft{ FEEDER_STAGE1_LEFT_MOTOR_ID };
    VictorSPX stage1MotorRight{ FEEDER_STAGE1_RIGHT_MOTOR_ID };
    VictorSPX stage2Motor{ FEEDER_STAGE2_MOTOR_ID };
#endif
  };
}  // namespace Rebels
