#include "Intake.h"
#include "Settings.h"
#include "Telemetry.h"

using namespace Rebels;

#define TELEMETRY_CATEGORY "Intake"

REGISTER_BOOL(IntakeVerboseDashboard, "Intake/Verbose", false)
REGISTER_DOUBLE(IntakeTimeToExtend, "Intake/Time To Extend", 0.5)
REGISTER_DOUBLE(IntakeRollerRamp, "Intake/Roller Ramp Time", 0.4)
REGISTER_DOUBLE(IntakeRollerPowerIn, "Intake/Roller Power In", 1.0)
REGISTER_DOUBLE(IntakeRollerPowerOut, "Intake/Roller Power Out", -1.0)

Intake::Intake() : StateMachine("Intake", IntakeStrings) {
  AddTransition(INTAKE_INIT, INTAKE_IDLE);
  AddTransition(INTAKE_IDLE, INTAKE_EXTENDING);
  AddTransition(INTAKE_EXTENDING, INTAKE_EXTENDED);
  AddTransition(INTAKE_EXTENDING, INTAKE_RETRACTING);
  AddTransition(INTAKE_EXTENDED, INTAKE_ROLLING);
  AddTransition(INTAKE_EXTENDED, INTAKE_RETRACTING);
  AddTransition(INTAKE_ROLLING, INTAKE_RETRACTING);
  AddTransition(INTAKE_RETRACTING, INTAKE_IDLE);
  AddTransition(INTAKE_RETRACTING, INTAKE_EXTENDING);
}

bool Intake::TransitionTo(IntakeState newState) {
  switch (newState) {
  case INTAKE_INIT:
  case INTAKE_IDLE:
    StopMotors();
    FireSolenoid(frc::DoubleSolenoid::kOff);
    return true;
  case INTAKE_EXTENDING:
    StopMotors();
    FireSolenoid(frc::DoubleSolenoid::kForward);
    return true;
  case INTAKE_EXTENDED:
    StopMotors();
    FireSolenoid(frc::DoubleSolenoid::kOff);
    return true;
  case INTAKE_ROLLING:
    SetRollerMotorPower(reversed ? GetIntakeRollerPowerOut() : GetIntakeRollerPowerIn());
    return true;
  case INTAKE_RETRACTING:
    StopMotors();
    FireSolenoid(frc::DoubleSolenoid::kReverse);
    return true;
  }
  return false;
}

void Intake::Update() {
#ifdef ENABLE_INTAKE
  rollerMotor.Set(ControlMode::PercentOutput, rollerMotorPower);
#endif

  switch (GetCurrentState()) {
  case INTAKE_INIT:
    Init();
    ForceState(INTAKE_IDLE);
    break;

  case INTAKE_IDLE:
    StopMotors();
    break;

  case INTAKE_EXTENDING:
    if (ElapsedSinceTransition() > units::second_t(GetIntakeTimeToExtend())) {
      ChangeState(INTAKE_EXTENDED);
    }
    break;

  case INTAKE_EXTENDED:
    if (ElapsedSinceTransition() > 0.1_s) {
      ChangeState(INTAKE_ROLLING);
    }
    break;

  case INTAKE_RETRACTING:
    if (ElapsedSinceTransition() > units::second_t(GetIntakeTimeToExtend())) {
      ChangeState(INTAKE_IDLE);
    }
    break;

  case INTAKE_ROLLING:
    break;
  }
}

void Intake::BackgroundUpdate() {
  if (!GetIntakeVerboseDashboard()) {
    return;
  }

  LOG_VALUE(TELEMETRY_CATEGORY, "Roller Motor Power", rollerMotorPower);
}

void Intake::Extend() {
  const auto state = GetCurrentState();
  if (state == INTAKE_IDLE || state == INTAKE_RETRACTING) {
    LOG_INFO(TELEMETRY_CATEGORY, "Extending");
    ChangeState(INTAKE_EXTENDING);
  }
}

void Intake::Retract() {
  const auto state = GetCurrentState();
  if (state != INTAKE_IDLE && state != INTAKE_RETRACTING) {
    LOG_INFO(TELEMETRY_CATEGORY, "Retracting");
    ChangeState(INTAKE_RETRACTING);
  }
}

void Intake::Grab() {
  if (reversed) {
    LOG_INFO(TELEMETRY_CATEGORY, "Grabbing");
    reversed = false;
  }
}

void Intake::Eject() {
  if (!reversed) {
    LOG_INFO(TELEMETRY_CATEGORY, "Ejecting");
    reversed = true;
  }
}

bool Intake::IsReversed() const {
  return reversed;
}

void Intake::Init() {
  LOG_INFO(TELEMETRY_CATEGORY, "Initializing");

#ifdef ENABLE_INTAKE
  const auto ramp = GetIntakeRollerRamp();
  rollerMotor.ConfigFactoryDefault();
  rollerMotor.ConfigClosedloopRamp(ramp);
  rollerMotor.ConfigOpenloopRamp(ramp);
#endif

  StopMotors();
}

void Intake::StopMotors() {
  SetRollerMotorPower(0);
}

void Intake::SetRollerMotorPower(double power) {
  if (power > 0.001 && rollerMotorPower < 0.001) {
    LOG_INFO(TELEMETRY_CATEGORY, "Initiating roller forward motion");
  }

  if (power < -0.001 && rollerMotorPower > -0.001) {
    LOG_INFO(TELEMETRY_CATEGORY, "Initiating roller backward motion");
  }

  rollerMotorPower = power;
}

void Intake::FireSolenoid(frc::DoubleSolenoid::Value value) {
#ifdef ENABLE_INTAKE
#ifdef ENABLE_COMPRESSOR
  solenoid.Set(value);
#endif
#endif
}
