#include "PointGuard.h"
#include "Settings.h"
#include "Telemetry.h"
#include "Limelight.h"
#include "Teleop.h"

using namespace Rebels;

#define TELEMETRY_CATEGORY "PointGuard"

REGISTER_BOOL(PointGuardVerboseDashboard, "PointGuard/Verbose", false)
REGISTER_DOUBLE(PointGuardMinShootTime, "PointGuard/Minimum Shoot Time", 1)
REGISTER_DOUBLE(PointGuardMinReadyTime, "PointGuard/Minimum Ready Time", 0.5)
REGISTER_STRING(PointGuardShooterValuesTable, "PointGuard/Shooter Values Table", "20:2750,10:3300,0:4000") /*15% angle limlight=20:2750,10:3300,0:4000*/
REGISTER_DOUBLE(DrivePIDValueP, "PointGuard/Drive PID/P", 0.01)
REGISTER_DOUBLE(DrivePIDValueI, "PointGuard/Drive PID/I", 0.0)
REGISTER_DOUBLE(DrivePIDValueD, "PointGuard/Drive PID/D", 0.001)
REGISTER_DOUBLE(RotationTargetValue, "PointGuard/Rotation Target", 0.0)
REGISTER_DOUBLE(RotationThreshold, "PointGuard/Rotation Threshold", 2.0)
REGISTER_DOUBLE(MaxRPM, "PointGuard/MaxRPM", 5000)

PointGuard::PointGuard() : StateMachine("PointGuard", PointGuardStrings) {
  AddTransition(PG_IDLE, PG_SHOOTING_VISION);
  AddTransition(PG_SHOOTING_VISION, PG_IDLE);
  AddTransition(PG_IDLE, PG_SHOOTING_MANUAL);
  AddTransition(PG_SHOOTING_MANUAL, PG_IDLE);
  SetUpdatePriority(MediumUpdateablePriority);

  // Intake and Feed to not make use of the state machine currently.

  intake = Intake::Get();
  feeder = Feeder::Get();
  shooter = Shooter::Get();
}

void PointGuard::Init() {
  LOG_INFO(TELEMETRY_CATEGORY, "Initializing");
  forceManualRPM = false;
  ejecting = false;
  intaking = false;
  shooting = false;
  shootingExpires = 0_s;
  Limelight::SetLedState(LimelightLedState::Off);

  // Load the table of shooter values.
  shooterValuesTable = ParseTable(GetPointGuardShooterValuesTable());
  if (shooterValuesTable.size() == 0) {
    LOG_ERROR(TELEMETRY_CATEGORY, "Could not load table. Forcing manual values.");
    forceManualRPM = true;
  }
  else {
    shooterValuesLogged = false;
  }
}

bool PointGuard::TransitionTo(PointGuardState newState) {

#ifdef ENABLE_DRIVE
  if (HasDrive()) {
    DisableDrive();
    teleop->EnableDrive();
  }
#endif

  switch (newState) {
  case PG_INIT:
    return true;
  case PG_IDLE:
    Limelight::SetLedState(Off);
    feeder->IsJogging(false);
    shooter->Stop();
    if (feeder->IsFeeding())
      feeder->Stop();
    return true;
  case PG_SHOOTING_MANUAL:
    shooter->Start();
    return true;
  case PG_SHOOTING_VISION:
    Limelight::SetLedState(On);
    pidController.Reset();
    pidController.SetPID(GetDrivePIDValueP(), GetDrivePIDValueI(), GetDrivePIDValueD());
    shooter->Start();
    return true;
  }
  return false;
}

void PointGuard::Update() {

  if (intaking && !feeder->IsIntaking()) {
    feeder->Intake();
  }
  else if (!intaking && feeder->IsIntaking())
  {
    feeder->Stop();
  }

  switch (GetCurrentState()) {
  case PG_INIT:
    Init();
    ForceState(PG_IDLE);
    break;
  case PG_IDLE:
    break;
  case PG_SHOOTING_MANUAL:
    if (shooter->IsReady() && !feeder->IsFeeding()) {
      feeder->Feed();
    }
    else if (!shooter->IsReady() && feeder->IsFeeding()) {
      feeder->Stop();
    }
    break;
  case PG_SHOOTING_VISION:
    if (Limelight::GetTv()) {
      // Limelight has a target, take over drive and keep it aimed at the target.
      double angleX = Limelight::GetTx();

#ifdef ENABLE_DRIVE
      if (!HasDrive()) {
        teleop->DisableDrive();
        EnableDrive();
      }

      // Calculate drive PID using limelight Tx value.
      double leftDrivePower = pidController.Calculate(angleX, GetRotationTargetValue());
      double rightDrivePower = -1 * leftDrivePower;
      (*drive)->DirectDrive(leftDrivePower, rightDrivePower);
#endif

      // Limelight has a target, update shooter speeds to match distance from target.
      double angleY = Limelight::GetTy();
      double rpm = 0;
      if (!forceManualRPM) {
        // Use the table of limelight->rpm values and interpolate between them to set the wheel speed.
        rpm = CalculateRPM(angleY, shooterValuesTable);
        shooter->SetFlywheelTarget(rpm); //std::clamp(rpm, 0.0, GetMaxRPM())
        shooter->SetTopRollerTarget(rpm);
      }


#ifdef ENABLE_DRIVE
      // If pointed at target.
      if (std::abs(angleX - GetRotationTargetValue()) < GetRotationThreshold()) {
#endif
        bool isInRpmRange = true;
        // If the requested RPM is above our max configured RPM the robot should move forwards.
        // if (rpm > GetMaxRPM()) {
        //   isInRpmRange = false;
        //   // Set up a PID to move forwards? (Baby steps my dude. Get the other stuff working first.)
        // }

        // if (rpm < GetMinRPM()) {
        //   isInRpmRange = false;
        //   // Set up PID to move backwards.
        // }

        if (shooter->IsReady() && isInRpmRange) {
          feeder->Feed();
        }
        else
        {
          if (feeder->IsFeeding())
            feeder->Stop();
        }
#ifdef ENABLE_DRIVE
      }
#endif

    }
    else {
#ifdef ENABLE_DRIVE
      // Give control back to the driver so that they can aim it towards the goal.
      if (HasDrive()) {
        DisableDrive();
        teleop->EnableDrive();
      }
#endif
    }
    break;
  }

  // synchronize intake/feeder behavior
  if (!ejecting) {
    if (intake->IsReversed()) {
      intake->Grab();
    }
  }
}

void PointGuard::BackgroundUpdate() {
  if (!GetPointGuardVerboseDashboard()) {
    return;
  }

  if (!shooterValuesLogged) {
    LOG_DEBUG(TELEMETRY_CATEGORY, fmt::format("Entries in table {}", shooterValuesTable.size()));
    std::string tableStr = "\n[Table]";
    std::map<double, double>::iterator iter;
    for (iter = shooterValuesTable.begin(); iter != shooterValuesTable.end(); iter++) {
      tableStr.append(fmt::format("\nKey: {} Val: {}", iter->first, iter->second));
    }
    LOG_DEBUG(TELEMETRY_CATEGORY, tableStr);
    shooterValuesLogged = true;
  }

  // LOG_VALUE(TELEMETRY_CATEGORY, "...", ...);
}

void PointGuard::StartIntake() {
  if (!intaking) {
    LOG_INFO(TELEMETRY_CATEGORY, "Starting intake");
    intaking = true;
    intake->Extend();
  }
}

void PointGuard::StopIntake() {
  if (intaking) {
    LOG_INFO(TELEMETRY_CATEGORY, "Stopping intake");
    intaking = false;
    intake->Retract();
  }
}

// Ejection is allowed with intake either extended or retracted.
// If intake happens to be extended we run it in reverse. At the same time
// we don't bother with its actual state, instead we instruct it to be
// in ejection mode, which would do nothing in retracted state.

void PointGuard::StartEjection() {
  if (!ejecting) {
    LOG_INFO(TELEMETRY_CATEGORY, "Starting ejection");
    ejecting = true;

    intake->Eject();
    feeder->Eject();
  }
}

void PointGuard::StopEjection() {
  if (ejecting) {
    LOG_INFO(TELEMETRY_CATEGORY, "Stopping ejection");
    ejecting = false;

    feeder->Stop();
  }
}

// Enaging/disengaging shooter only impacts the shooter wheels, to actually
// shoot a ball the feeder must be instructed to actively feed balls to
// the shooter. As such it is ok to engage the shooter regardless of what
// other subsystems in the pipeline are doing.

void PointGuard::StartShootingManual() {
  LOG_INFO(TELEMETRY_CATEGORY, "Starting shooting manually");
  if (GetCurrentState() != PG_SHOOTING_MANUAL) {
    ChangeState(PG_SHOOTING_MANUAL);
  }
}

void PointGuard::StartShootingVision() {
  LOG_INFO(TELEMETRY_CATEGORY, "Starting shooting by vision");
  if (GetCurrentState() != PG_SHOOTING_VISION) {
    ChangeState(PG_SHOOTING_VISION);
  }
}

void PointGuard::StopShooting() {
  LOG_INFO(TELEMETRY_CATEGORY, "Stopping shooting");
  if (GetCurrentState() != PG_IDLE) {
    ChangeState(PG_IDLE);
  }
}

std::map<double, double> PointGuard::ParseTable(std::string str) {
  std::map<double, double> map;
  std::stringstream stream(str);
  std::string pair = "";

  try {
    while (std::getline(stream, pair, ',')) {
      const auto index = pair.find(':');
      double key = std::stod(pair.substr(0, index));
      double val = std::stod(pair.substr(index + 1, pair.size()));
      map[key] = val;
    }
  }
  catch (std::exception ex) {
    map.clear();
    return map;
  }

  return map;
}

double PointGuard::CalculateRPM(double inputMeasurement, std::map<double, double> rpmTable) {
  if (rpmTable.size() < 2)
    return 0;

  double firstKey = 0;
  double firstVal = 0;
  double secondKey = 0;
  double secondVal = 0;

  std::map<double, double>::iterator iter;
  for (iter = rpmTable.begin(); iter != rpmTable.end(); iter++) {
    // If the value is in our table, return it. No need to calculate anything.
    if (inputMeasurement == iter->first)
      return iter->second;

    // If the measurment is less than the first value, we can caluclate it using the first two values of the list.
    if (iter == rpmTable.begin() && inputMeasurement < iter->first) {
      firstKey = iter->first;
      firstVal = iter->second;
      iter++;
      secondKey = iter->first;
      secondVal = iter->second;
      break;
    }

    // If the measurment is greater than the last value, we can calculate it using the last and second last value of the list.
    if (iter == --rpmTable.end() && inputMeasurement > iter->first) {
      secondKey = iter->first;
      secondVal = iter->second;
      break;
    }

    // Else, we continually store the smaller value in the first pair until we find a pair that is larger.
    if (inputMeasurement < iter->first) {
      secondKey = iter->first;
      secondVal = iter->second;
      break;
    }
    else {
      firstKey = iter->first;
      firstVal = iter->second;
    }
  }

  // Calculate the interpolation between the two values from our table and return.
  // y0(x1-x)+y1(x-x0)
  // _________________
  //      x1-x0
  return (firstVal * (secondKey - inputMeasurement) + secondVal * (inputMeasurement - firstKey)) / (secondKey - firstKey);
}

void PointGuard::EnableDrive() {
#ifdef ENABLE_DRIVE
  LOG_INFO(TELEMETRY_CATEGORY, "Taking over the drive system");
  if (drive.get() == nullptr) {
    drive.reset(new Drive(SystemName));
  }
  (*drive)->SetBrakeMode();
#endif
}

void PointGuard::DisableDrive() {
#ifdef ENABLE_DRIVE
  if (drive.get() != nullptr) {
    LOG_INFO(TELEMETRY_CATEGORY, "Releasing the drive system");
    drive.reset(nullptr);
  }
#endif
}

bool PointGuard::HasDrive() {
  return (drive.get() != nullptr && (*drive)->GetOwner() == SystemName);
}

void PointGuard::SetTeleopObj(Teleop& tele) {
  teleop = &tele;
}
