#pragma once

#include <frc2/command/PIDSubsystem.h>

#include "Constants.h"
#include "Settings.h"
#include "Singleton.h"
#include "StateMachine.h"
#include "Intake.h"
#include "Feeder.h"
#include "Shooter.h"
#include "Drive.h"
#include "Teleop.h"

namespace Rebels {

  enum PointGuardState {
    PG_INIT,
    PG_IDLE,
    PG_SHOOTING_MANUAL,
    PG_SHOOTING_VISION
  };

  const StateMachine<PointGuardState>::StateStrings PointGuardStrings = {
    "Init",
    "Idle",
    "Shooting_Manual",
    "Shooting_Vision"
  };

  class PointGuard : public StateMachine<PointGuardState> {
    DECLARE_SINGLETON(PointGuard)

  public:
    virtual ~PointGuard() = default;
    void Reset() { ForceState(PG_INIT); }

  public:
    virtual void Update() override;
    virtual void BackgroundUpdate() override;

    // Actions
    void StartIntake();
    void StopIntake();
    void StartEjection();
    void StopEjection();
    void StartShootingManual();
    void StartShootingVision();
    void StopShooting();

    virtual void EnableDrive();
    virtual void DisableDrive();
    bool HasDrive();

    void SetTeleopObj(Teleop& tele);

  protected:
    virtual bool TransitionTo(PointGuardState newState) override;
    static const constexpr std::string_view SystemName{ "PointGuard" };

  private:
    PointGuard();
    void Init();
    double CalculateRPM(double inputMeasurement, std::map<double, double> rpmTable);
    std::map<double, double> ParseTable(std::string str);

  private:
    std::atomic_bool ejecting = false;
    std::atomic_bool intaking = false;
    std::atomic_bool shooting = false;
    units::second_t shootingExpires{ 0.0 };
    units::second_t shooterLastReadyAt{ 0.0 };
    std::map<double, double> shooterValuesTable;
    bool forceManualRPM = false;
    bool shooterValuesLogged = false;

    Intake* intake = nullptr;
    Feeder* feeder = nullptr;
    Shooter* shooter = nullptr;
    Teleop* teleop = nullptr;
    std::unique_ptr<Rebels::Drive> drive;

    frc2::PIDController pidController{ 0,0,0 };
  };
}  // namespace Rebels
