#include <frc/DriverStation.h>

#include "Robot.h"
#include "StateMachines.h"
#include "Settings.h"
#include "Constants.h"
#include "Telemetry.h"

#include "Intake.h"
#include "Feeder.h"
#include "Shooter.h"
#include "PointGuard.h"
#include "Climber.h"
#include "Limelight.h"

using namespace Rebels;

#define TELEMETRY_CATEGORY "Robot"

Robot::Robot() {}

void Robot::RobotInit() {
  LOG_INFO(TELEMETRY_CATEGORY, "Initializing major systems");
  // [HERE] add initialization code for major subsystems
  // e.g. Intake::Create();
  Drive::Init();
  Intake::Create();
  Feeder::Create();
  Shooter::Create();
  PointGuard::Create();
  Climber::Create();

  Limelight::SetLedState(LimelightLedState::Off);

  LOG_INFO(TELEMETRY_CATEGORY, "Creating autonomous system");
  autonomous.reset(new Autonomous());

  LOG_INFO(TELEMETRY_CATEGORY, "Creating teleop system");
  teleop.reset(new Teleop());

  PointGuard::Get()->SetTeleopObj(*teleop);
}

void Robot::RobotPeriodic() {}

void Robot::AutonomousInit() {
  teleop->DisableDrive();
  autonomous->Start();
}

void Robot::AutonomousPeriodic() {
  // disable auto and give teleop system control
  // this is only applicable during testing, not real match
  if (teleop->IsJoystickMoved()) {
    LOG_INFO(TELEMETRY_CATEGORY, "Driver moved stick, disabling auto");
    autonomous->Stop();
    teleop->EnableDrive();
  }

  if (autonomous->IsRunning()) {
    autonomous->Periodic();
  }
  else {
    teleop->Periodic();
  }
}

void Robot::TeleopInit() {
  if (autonomous.get() != nullptr && autonomous->IsRunning()) {
    autonomous->Stop();
  }
  teleop->EnableDrive();
}

void Robot::TeleopPeriodic() {
  teleop->Periodic();
}

void Robot::EnabledInit() {
  teleop->ClearButtonState();
}

void Robot::DisabledInit() {
  // Perform any resetting of state machines or other systems
  // that should only happen when we're testing the manual DS control modes,
  // not during a real match.
  LOG_WARNING(TELEMETRY_CATEGORY, "Resetting state machines after testing");

  // [HERE] add reset code for major subsystems
  Intake::Get()->Reset();
  Feeder::Get()->Reset();
  Shooter::Get()->Reset();
  Climber::Get()->Reset();

  if (autonomous.get() != nullptr && autonomous->IsRunning()) {
    autonomous->Stop();
  }
}

void Robot::DisabledPeriodic() {
  if (TimeSinceDisabled() > 2_s) {
    if (autonomous.get() != nullptr && autonomous->IsRunning()) {
      LOG_INFO(TELEMETRY_CATEGORY, "Stopping auto after being disabled for 2 seconds");
      autonomous->Stop();
      teleop->EnableDrive();
    }
  }
}

#ifndef RUNNING_FRC_TESTS
int main() {
  return frc::StartRobot<Robot>();
}
#endif