#include "Shooter.h"
#include "Settings.h"
#include "Telemetry.h"

using namespace Rebels;

#define TELEMETRY_CATEGORY "Shooter"

// TalonFX getSelectedSensorVelocity returns position units per 100ms
// where one revolution corresponds to 2048 position units
const double TALONFX_UNITS_PER_REVOLUTION = 2048.0;
const double TALONFX_VELOCITY_TO_RPM = 600.0 / TALONFX_UNITS_PER_REVOLUTION;
const double TALONFX_RPM_TO_VELOCITY = 1.0 / TALONFX_VELOCITY_TO_RPM;

REGISTER_BOOL(ShooterVerboseDashboard, "Shooter/Verbose", false)

REGISTER_DOUBLE(FlywheelManualRPM, "Shooter/Flywheel/Manual RPM", 2400)
REGISTER_DOUBLE(FlywheelTargetRpm, "Shooter/Flywheel/Target RPM", 0)
REGISTER_DOUBLE(FlywheelCurrentRpm, "Shooter/Flywheel/Current RPM", 0)
REGISTER_DOUBLE(FlywheelP, "Shooter/Flywheel/PID/P", 0.4)
REGISTER_DOUBLE(FlywheelI, "Shooter/Flywheel/PID/I", 0)
REGISTER_DOUBLE(FlywheelD, "Shooter/Flywheel/PID/D", 0)
REGISTER_DOUBLE(FlywheelF, "Shooter/Flywheel/PID/F", 0.05)
REGISTER_DOUBLE(FlywheelLoopError, "Shooter/Flywheel/PID - Error %", 2)
REGISTER_DOUBLE(FlywheelRamp, "Shooter/Flywheel/Ramp Time", 1)
REGISTER_DOUBLE(FlywheelThresholdRpm, "Shooter/Flywheel/Threshold RPM", 100)

REGISTER_DOUBLE(TopRollerManualRPM, "Shooter/TopRoller/Manual RPM", 2400)
REGISTER_DOUBLE(TopRollerTargetRpm, "Shooter/TopRoller/Target RPM", 0)
REGISTER_DOUBLE(TopRollerCurrentRpm, "Shooter/TopRoller/Current RPM", 0)
REGISTER_DOUBLE(TopRollerP, "Shooter/TopRoller/PID/P", 0.4)
REGISTER_DOUBLE(TopRollerI, "Shooter/TopRoller/PID/I", 0)
REGISTER_DOUBLE(TopRollerD, "Shooter/TopRoller/PID/D", 0)
REGISTER_DOUBLE(TopRollerF, "Shooter/TopRoller/PID/F", 0.05)
REGISTER_DOUBLE(TopRollerLoopError, "Shooter/TopRoller/PID - Error %", 2)
REGISTER_DOUBLE(TopRollerRamp, "Shooter/TopRoller/Ramp Time", 1)
REGISTER_DOUBLE(TopRollerThresholdRpm, "Shooter/TopRoller/Threshold RPM", 100)

REGISTER_BOOL(IsShooterReadyToShoot, "Shooter/IsShooterReady", false)

Shooter::Shooter() : StateMachine("Shooter", ShooterStrings) {
  AddTransition(SHOOTER_SHOOTING, SHOOTER_IDLE);
  AddTransition(SHOOTER_IDLE, SHOOTER_SHOOTING);
}

bool Shooter::TransitionTo(ShooterState newState) {
  switch (newState) {
  case SHOOTER_INIT:
  case SHOOTER_IDLE:
    SetShooterSpeedRpm(0, 0);
    return true;
  case SHOOTER_SHOOTING:
    SetTopRollerTargetRpm(GetTopRollerManualRPM());
    SetFlywheelTargetRpm(GetFlywheelManualRPM());
    SetShooterSpeedRpm(GetFlywheelTargetRpm(), GetTopRollerTargetRpm());
    return true;
  }

  return false;
}

void Shooter::Update() {

  SetIsShooterReadyToShoot(IsReady());

#ifdef ENABLE_SHOOTER
  flywheelCurrentRpm = flywheelMotor.GetSelectedSensorVelocity(0) * TALONFX_VELOCITY_TO_RPM;
  topRollerCurrentRpm = topRollerMotor.GetSelectedSensorVelocity(0) * TALONFX_VELOCITY_TO_RPM;
#else
  flywheelCurrentRpm = 0;
  topRollerCurrentRpm = 0;
#endif

  SetFlywheelCurrentRpm(flywheelCurrentRpm);
  SetTopRollerCurrentRpm(topRollerCurrentRpm);

  switch (GetCurrentState()) {
  case SHOOTER_INIT:
    Init();
    ForceState(SHOOTER_IDLE);
    break;

  case SHOOTER_IDLE:
    break;

  case SHOOTER_SHOOTING:
    SetShooterSpeedRpm(GetFlywheelTargetRpm(), GetTopRollerTargetRpm());
    break;
  }
}

void Shooter::BackgroundUpdate() {
  if (!GetShooterVerboseDashboard()) {
    return;
  }

#ifdef ENABLE_SHOOTER
  LOG_VALUE(TELEMETRY_CATEGORY, "Flywheel/Current RPM", flywheelCurrentRpm);
  LOG_VALUE(TELEMETRY_CATEGORY, "Flywheel/Target RPM", flywheelDesiredRpm);
  LOG_VALUE(TELEMETRY_CATEGORY, "Flywheel/Closed Loop Error %", 100.0 * flywheelMotor.GetClosedLoopError(0) / flywheelMotor.GetClosedLoopTarget(0));
  LOG_VALUE(TELEMETRY_CATEGORY, "Flywheel/Supply Current (A)", flywheelMotor.GetSupplyCurrent());
  LOG_VALUE(TELEMETRY_CATEGORY, "Flywheel/Stator Current (A)", flywheelMotor.GetStatorCurrent());

  LOG_VALUE(TELEMETRY_CATEGORY, "TopRoller/Current RPM", topRollerCurrentRpm);
  LOG_VALUE(TELEMETRY_CATEGORY, "TopRoller/Target RPM", topRollerDesiredRpm);
  LOG_VALUE(TELEMETRY_CATEGORY, "TopRoller/Closed Loop Error %", 100.0 * topRollerMotor.GetClosedLoopError(0) / topRollerMotor.GetClosedLoopTarget(0));
  LOG_VALUE(TELEMETRY_CATEGORY, "TopRoller/Supply Current (A)", topRollerMotor.GetSupplyCurrent());
  LOG_VALUE(TELEMETRY_CATEGORY, "TopRoller/Stator Current (A)", topRollerMotor.GetStatorCurrent());
#endif
}



void Shooter::Start() {
  if (GetCurrentState() == SHOOTER_IDLE) {
    LOG_INFO(TELEMETRY_CATEGORY, "Starting");
    ChangeState(SHOOTER_SHOOTING);
  }
}

void Shooter::Stop() {
  if (GetCurrentState() == SHOOTER_SHOOTING) {
    LOG_INFO(TELEMETRY_CATEGORY, "Stopping");
    ChangeState(SHOOTER_IDLE);
  }
}

bool Shooter::IsRunning() const {
  return (GetCurrentState() == SHOOTER_SHOOTING);
}

bool Shooter::IsReady() const {
  const auto thresholdFlywheelRpm = GetFlywheelThresholdRpm();
  const auto thresholdTopRollerRpm = GetTopRollerThresholdRpm();
  const auto ready = IsRunning() &&
    std::abs(flywheelCurrentRpm - GetFlywheelTargetRpm()) < thresholdFlywheelRpm &&
    std::abs(topRollerCurrentRpm - GetTopRollerTargetRpm()) < thresholdTopRollerRpm;

  /*LOG_DEBUG(TELEMETRY_CATEGORY, fmt::format("Ready = {}, Flywheel = {:.0f} [# {:.0f}], TopRoller = {:.0f} [# {:.0f}]",
    ready, flywheelCurrentRpm, thresholdFlywheelRpm, topRollerCurrentRpm, thresholdTopRollerRpm));*/
  return ready;
}

void Shooter::SetFlywheelTarget(double rpm) {
  SetFlywheelTargetRpm(rpm);
}

void Shooter::SetTopRollerTarget(double rpm) {
  SetTopRollerTargetRpm(rpm);
}

void Shooter::MakeSound() {
  if (GetCurrentState() == SHOOTER_IDLE) {
    LOG_INFO(TELEMETRY_CATEGORY, "Making noise");
    if (IsMusicPlaying())
      StopMusic();
    PlayMusic();
  }
  else {
    LOG_INFO(TELEMETRY_CATEGORY, fmt::format("Could not make sound as shooter is currently in state:{0}", ShooterStrings[GetCurrentState()]));
  }
}

void Shooter::Init() {
  LOG_INFO(TELEMETRY_CATEGORY, "Initializing");
  flywheelDesiredRpm = GetFlywheelManualRPM();
  topRollerDesiredRpm = GetTopRollerManualRPM();
#ifdef ENABLE_SHOOTER
  InitMotor(flywheelMotor, true);
  InitMotor(topRollerMotor, false);
  InitOrchestra();
#endif
}

void Shooter::InitMotor(TalonFX& motor, bool bInverted) {
  const auto timeout = TIMEOUT.convert<units::ms>().value();

#ifdef ENABLE_SHOOTER
  motor.ConfigFactoryDefault();
  motor.ConfigNominalOutputForward(0, timeout);
  motor.ConfigNominalOutputReverse(0, timeout);
  motor.ConfigPeakOutputForward(1, timeout);
  motor.ConfigPeakOutputReverse(-1, timeout);
  motor.SetInverted(bInverted);
  motor.SetNeutralMode(motorcontrol::NeutralMode::Coast);

  ctre::phoenix::motorcontrol::SupplyCurrentLimitConfiguration limitConfig;

  limitConfig.enable = true;
  limitConfig.currentLimit = 40;            // Amps
  limitConfig.triggerThresholdCurrent = 80; // Amps
  limitConfig.triggerThresholdTime = 2;     // seconds

  motor.ConfigSupplyCurrentLimit(limitConfig);
#endif
}

void Shooter::InitOrchestra() {
#ifdef ENABLE_SHOOTER
  orchestra.AddInstrument(flywheelMotor);
  orchestra.AddInstrument(topRollerMotor);
  orchestra.LoadMusic("song2.chrp");
#endif
}

void Shooter::SetMotorSpeedRpm(TalonFX& motor, const double rpm, const double kP, const double kI, const double kD, const double kF,
  const double ramp, const double relativeError) {
#ifdef ENABLE_SHOOTER
  if (rpm <= 0) {
    motor.Set(ControlMode::PercentOutput, 0);
  }
  else {
    const auto speed = rpm * TALONFX_RPM_TO_VELOCITY;
    const auto error = speed * relativeError / 100.0;

    motor.ConfigClosedloopRamp(ramp);
    motor.ConfigOpenloopRamp(ramp);
    motor.Config_kP(0, kP);
    motor.Config_kI(0, kI);
    motor.Config_kD(0, kD);
    motor.Config_kF(0, kF);
    motor.ConfigAllowableClosedloopError(0, error);
    motor.Set(ControlMode::Velocity, speed);
  }
#endif
}

void Shooter::SetShooterSpeedRpm(const double flywheelRPM, const double topRollerRPM) {
#ifdef ENABLE_SHOOTER
  SetMotorSpeedRpm(flywheelMotor, flywheelRPM, GetFlywheelP(), GetFlywheelI(), GetFlywheelD(), GetFlywheelF(), GetFlywheelRamp(), GetFlywheelLoopError());
  SetMotorSpeedRpm(topRollerMotor, topRollerRPM, GetTopRollerP(), GetTopRollerI(), GetTopRollerD(), GetTopRollerF(), GetTopRollerRamp(), GetTopRollerLoopError());
#endif
}

void Shooter::PlayMusic() {
#ifdef ENABLE_SHOOTER
  orchestra.Play();
#endif
}

void Shooter::StopMusic() {
#ifdef ENABLE_SHOOTER
  orchestra.Stop();
#endif
}

bool Shooter::IsMusicPlaying() {
#ifdef ENABLE_SHOOTER
  return orchestra.IsPlaying();
#endif
  return false;
}